# -*- coding: utf-8 -*-
"""Data access modules for loading system-of-systems model configuration
"""

# import classes for access like ::
#         from smif.data_layer.file import YamlConfigStore`
from smif.data_layer.file.file_config_store import YamlConfigStore
from smif.data_layer.file.file_data_store import (CSVDataStore,
                                                  ParquetDataStore)
from smif.data_layer.file.file_metadata_store import FileMetadataStore

# Define what should be imported as * ::
#         from smif.data_layer.file import *
__all__ = ['CSVDataStore', 'FileMetadataStore',
           'ParquetDataStore', 'YamlConfigStore']



"""The decision module handles the three planning levels

Currently, only pre-specified planning is implemented.

The choices made in the three planning levels influence the set of interventions
and assets available within a model run.

The interventions available in a model run are stored in a dict keyed by name.

"""
__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"

import itertools
import os
from abc import ABCMeta, abstractmethod
from logging import getLogger

from smif.data_layer.data_handle import ResultsHandle
from smif.data_layer.model_loader import ModelLoader
from smif.exception import SmifDataNotFoundError


class DecisionManager(object):
    """A DecisionManager is initialised with one or more model run strategies
    that refer to DecisionModules such as pre-specified planning,
    a rule-based models or multi-objective optimisation.
    These implementations influence the combination and ordering of decision
    iterations and model timesteps that need to be performed by the model runner.

    The DecisionManager presents a simple decision loop interface to the model runner,
    in the form of a generator which allows the model runner to iterate over the
    collection of independent simulations required at each step.

    The DecisionManager collates the output of the decision algorithms and
    writes the post-decision state through a ResultsHandle. This allows Models
    to access a given decision state (identified uniquely by timestep and
    decision iteration id).

    The :py:meth:`get_decisions` method passes a ResultsHandle down to a
    DecisionModule, allowing the DecisionModule to access model results from
    previous timesteps and decision iterations when making decisions

    Arguments
    ---------
    store: smif.data_layer.store.Store
    """

    def __init__(self, store, timesteps, modelrun_name, sos_model):

        self.logger = getLogger(__name__)

        self._store = store
        self._modelrun_name = modelrun_name
        self._sos_model = sos_model
        self._timesteps = timesteps
        self._decision_module = None

        self._register = {}
        for sector_model in sos_model.sector_models:
            self._register.update(self._store.read_interventions(sector_model.name))
        self.planned_interventions = {}

        strategies = self._store.read_strategies(modelrun_name)
        self.logger.info("%s strategies found", len(strategies))
        self.pre_spec_planning = self._set_up_pre_spec_planning(modelrun_name, strategies)
        self._set_up_decision_modules(modelrun_name, strategies)

    def _set_up_pre_spec_planning(self, modelrun_name, strategies):

        pre_spec_planning = None

        # Read in the historical interventions (initial conditions) directly
        initial_conditions = self._store.read_all_initial_conditions(modelrun_name)

        # Read in strategies
        planned_interventions = []
        planned_interventions.extend(initial_conditions)

        for index, strategy in enumerate(strategies):
            # Extract pre-specified planning interventions
            if strategy['type'] == 'pre-specified-planning':

                msg = "Adding %s planned interventions to pre-specified-planning %s"
                self.logger.info(msg, len(strategy['interventions']), index)

                planned_interventions.extend(strategy['interventions'])

        # Create a Pre-Specified planning decision module with all
        # the planned interventions
        if planned_interventions:
            pre_spec_planning = PreSpecified(self._timesteps,
                                             self._register,
                                             planned_interventions)
            self.planned_interventions = {x['name'] for x in planned_interventions}

        return pre_spec_planning

    def _set_up_decision_modules(self, modelrun_name, strategies):

        strategy_types = [x['type'] for x in strategies
                          if x['type'] != 'pre-specified-planning']
        if len(set(strategy_types)) > 1:
            msg = "Cannot use more the 2 type of strategy simultaneously"
            raise NotImplementedError(msg)

        for strategy in strategies:

            if strategy['type'] != 'pre-specified-planning':

                loader = ModelLoader()

                # absolute path to be crystal clear for ModelLoader when loading python class
                strategy['path'] = os.path.normpath(
                    os.path.join(self._store.model_base_folder, strategy['path']))
                strategy['timesteps'] = self._timesteps
                # Pass a reference to the register of interventions
                strategy['register'] = self.available_interventions

                strategy['name'] = strategy['classname'] + '_' + strategy['type']

                self.logger.debug("Trying to load strategy: %s", strategy['name'])
                decision_module = loader.load(strategy)
                self._decision_module = decision_module

    @property
    def available_interventions(self):
        """Returns a register of available interventions, i.e. those not planned
        """
        edited_register = {name: self._register[name]
                           for name in self._register.keys() -
                           self.planned_interventions}
        return edited_register

    def get_intervention(self, value):
        try:
            return self._register[value]
        except KeyError:
            msg = ""
            raise SmifDataNotFoundError(msg.format(value))

    def decision_loop(self):
        """Generate bundles of simulation steps to run

        Each call to this method returns a dict:

            {
                'decision_iterations': list of decision iterations (int),
                'timesteps': list of timesteps (int),
                'decision_links': (optional) dict of {
                    decision iteration in current bundle: decision iteration of previous bundle
                }
            }

        A bundle is composed differently according to the implementation of the
        contained DecisionModule.  For example:

        With only pre-specified planning, there is a single step in the loop,
        with a single decision iteration with timesteps covering the entire model horizon.

        With a rule based approach, there might be many steps in the loop, each with a single
        decision iteration and single timestep, moving on once some threshold is satisfied.

        With a genetic algorithm, there might be a configurable number of steps in the loop,
        each with multiple decision iterations (one for each member of the algorithm's
        population) and timesteps covering the entire model horizon.

        Implicitly, if the bundle returned in an iteration contains multiple decision
        iterations, they can be performed in parallel. If each decision iteration contains
        multiple timesteps, they can also be parallelised, so long as there are no temporal
        dependencies.

        Decision links are only required if the bundle timesteps do not start from the first
        timestep of the model horizon.
    """
        self.logger.debug("Calling decision loop")
        if self._decision_module:
            while True:
                bundle = next(self._decision_module)
                if bundle is None:
                    break
                self.logger.debug("Bundle returned: %s", bundle)
                self._get_and_save_bundle_decisions(bundle)
                yield bundle
        else:
            bundle = {
                'decision_iterations': [0],
                'timesteps': [x for x in self._timesteps]
            }
            self._get_and_save_bundle_decisions(bundle)
            yield bundle

    def _get_and_save_bundle_decisions(self, bundle):
        """Iterate over bundle and write decisions

        Arguments
        ---------
        bundle : dict

        Returns
        -------
        bundle : dict
            One definitive bundle across the decision modules

        """
        for iteration, timestep in itertools.product(
                bundle['decision_iterations'],
                bundle['timesteps']):
            self.get_and_save_decisions(iteration, timestep)

    def get_and_save_decisions(self, iteration, timestep):
        """Retrieves decisions for given timestep and decision iteration from each decision
        module and writes them to the store as state.

        Calls each contained DecisionModule for the given timestep and decision iteration in
        the `data_handle`, retrieving a list of decision dicts (keyed by intervention name and
        build year).

        These decisions are then written to a state file using the data store.

        Arguments
        ---------
        timestep : int
        iteration : int
        """
        results_handle = ResultsHandle(
            store=self._store,
            modelrun_name=self._modelrun_name,
            sos_model=self._sos_model,
            current_timestep=timestep,
            timesteps=self._timesteps,
            decision_iteration=iteration
        )
        decisions = []
        if self._decision_module:
            decisions.extend(self._decision_module.get_decision(results_handle))

        if self.pre_spec_planning:
            decisions.extend(self.pre_spec_planning.get_decision(results_handle))

        self.logger.debug(
            "Retrieved %s decisions from %s", len(decisions), str(self._decision_module))

        self.logger.debug(
            "Writing state for timestep %s and interation %s", timestep, iteration)

        self._store.write_state(decisions, self._modelrun_name, timestep, iteration)


class DecisionModule(metaclass=ABCMeta):
    """Abstract class which provides the interface to user defined decision modules.

    These mechanisms could include a Rule-based Approach or Multi-objective Optimisation.

    This class provides two main methods, ``__next__`` which is normally
    called implicitly as a call to the class as an iterator, and ``get_decision()``
    which takes as arguments a smif.data_layer.data_handle.ResultsHandle object.

    Arguments
    ---------
    timesteps : list
        A list of planning timesteps
    register : dict
        Reference to a dict of iterventions
    """

    """Current iteration of the decision module
    """
    def __init__(self, timesteps, register):
        self.timesteps = timesteps
        self._register = register
        self.logger = getLogger(__name__)

    def __next__(self):
        return self._get_next_decision_iteration()

    @property
    def interventions(self):
        """Return the list of available interventions

        Returns
        -------
        list
        """
        return self._register

    def get_intervention(self, name):
        """Return an intervention dict

        Returns
        -------
        dict
        """
        try:
            return self._register[name]
        except KeyError:
            msg = "Intervention '{}' is not found in the list of available interventions"
            raise SmifDataNotFoundError(msg.format(name))

    @abstractmethod
    def _get_next_decision_iteration(self):
        """Implement to return the next decision iteration

        Within a list of decision-iteration/timestep pairs, the assumption is
        that all decision iterations can be run in parallel
        (otherwise only one will be returned) and within a decision interation,
        all timesteps may be run in parallel as long as there are no
        inter-timestep state dependencies

        Returns
        -------
        dict
            Yields a dictionary keyed by ``decision iterations`` whose values
            contain a list of iteration integers, ``timesteps`` whose values
            contain a list of timesteps run in each decision iteration and the
            optional ``decision_links`` which link the decision interation of the
            current bundle to that of the previous bundle::

            {
                'decision_iterations': list of decision iterations (int),
                'timesteps': list of timesteps (int),
                'decision_links': (optional) dict of {
                    decision iteration in current bundle: decision iteration of previous bundle
                }
            }
        """
        raise NotImplementedError

    @abstractmethod
    def get_decision(self, results_handle):
        """Return decisions for a given timestep and decision iteration

        Parameters
        ----------
        results_handle : smif.data_layer.data_handle.ResultsHandle

        Returns
        -------
        list of dict

        Examples
        --------
        >>> register = {'intervention_a': {'capital_cost': {'value': 1234}}}
        >>> dm = DecisionModule([2010, 2015], register)
        >>> dm.get_decision(results_handle)
        [{'name': 'intervention_a', 'build_year': 2010}])
        """
        raise NotImplementedError


class PreSpecified(DecisionModule):
    """Pre-specified planning

    Parameters
    ----------
    timesteps : list
        A list of the timesteps included in the model horizon
    register : dict
        A dict of intervention dictionaries keyed by unique intervention name
    planned_interventions : list
        A list of dicts ``{'name': 'intervention_name', 'build_year': 2010}``
        representing historical or planned interventions
    """
    def __init__(self, timesteps, register, planned_interventions):
        super().__init__(timesteps, register)
        self._planned = planned_interventions

    def _get_next_decision_iteration(self):
        return {
            'decision_iterations': [0],
            'timesteps': [x for x in self.timesteps]
        }

    def get_decision(self, results_handle):
        """Return a dict of intervention names built in timestep

        Arguments
        ---------
        results_handle : smif.data_layer.data_handle.ResultsHandle
            A reference to a smif results handle

        Returns
        -------
        list of dict

        Examples
        --------
        >>> dm = PreSpecified([2010, 2015], register,
        [{'name': 'intervention_a', 'build_year': 2010}])
        >>> dm.get_decision(results_handle)
        [{'name': intervention_a', 'build_year': 2010}]
        """
        decisions = []
        timestep = results_handle.current_timestep

        assert isinstance(self._planned, list)

        for intervention in self._planned:
            build_year = int(intervention['build_year'])

            data = self._register[intervention['name']]
            lifetime = data['technical_lifetime']['value']

            if self.buildable(build_year, timestep) and \
               self.within_lifetime(build_year, timestep, lifetime):
                decisions.append(intervention)
        return decisions

    def buildable(self, build_year, timestep):
        """Interventions are deemed available if build_year is less than next timestep

        For example, if `a` is built in 2011 and timesteps are
        [2005, 2010, 2015, 2020] then buildable returns True for timesteps
        2010, 2015 and 2020 and False for 2005.
        """
        if not isinstance(build_year, (int, float)):
            msg = "Build Year should be an integer but is a {}"
            raise TypeError(msg.format(type(build_year)))
        if timestep not in self.timesteps:
            raise ValueError("Timestep not in model timesteps")
        index = self.timesteps.index(timestep)
        if index == len(self.timesteps) - 1:
            next_year = timestep + 1
        else:
            next_year = self.timesteps[index + 1]

        if int(build_year) < next_year:
            return True
        else:
            return False

    def within_lifetime(self, build_year, timestep, lifetime):
        """Interventions are deemed active if build_year + lifetime >= timestep

        Arguments
        ---------
        build_year : int
        timestep : int
        lifetime : int
        """
        if not isinstance(build_year, (int, float)):
            msg = "Build Year should be an integer but is a {}"
            raise TypeError(msg.format(type(build_year)))

        try:
            build_year = int(build_year)
        except ValueError:
            raise ValueError(
                "A build year must be a valid integer. Received {}.".format(build_year))

        try:
            lifetime = int(lifetime)
        except ValueError:
            lifetime = float("inf")
        if lifetime < 0:
            msg = "The value of lifetime cannot be negative"
            raise ValueError(msg)
        if timestep <= build_year + lifetime:
            return True
        else:
            return False


class RuleBased(DecisionModule):
    """Rule-base decision modules
    """

    def __init__(self, timesteps, register):
        super().__init__(timesteps, register)
        self.satisfied = False
        self.current_timestep = self.first_timestep
        self.current_iteration = 0
        # keep internal account of max iteration reached per timestep
        self._max_iteration_by_timestep = {self.first_timestep: 0}
        self.logger = getLogger(__name__)

    @property
    def next_timestep(self):
        index_current_timestep = self.timesteps.index(self.current_timestep)
        try:
            return self.timesteps[index_current_timestep + 1]
        except IndexError:
            return None

    @property
    def previous_timestep(self):
        index_current_timestep = self.timesteps.index(self.current_timestep)
        if index_current_timestep > 0:
            return self.timesteps[index_current_timestep - 1]
        else:
            return None

    @property
    def first_timestep(self):
        return min(self.timesteps)

    @property
    def last_timestep(self):
        return max(self.timesteps)

    def _get_next_decision_iteration(self):
        if self.satisfied and (self.current_timestep == self.last_timestep):
            return None
        elif self.satisfied and (self.current_timestep < self.last_timestep):
            self._max_iteration_by_timestep[self.current_timestep] = \
                self.current_iteration
            self.satisfied = False
            self.current_timestep = self.next_timestep
            self.current_iteration += 1
            return self._make_bundle()
        else:
            self.current_iteration += 1
            return self._make_bundle()

    def _make_bundle(self):
        bundle = {'decision_iterations': [self.current_iteration],
                  'timesteps': [self.current_timestep]}

        if self.current_timestep > self.first_timestep:
            bundle['decision_links'] = {
                self.current_iteration: self._max_iteration_by_timestep[
                    self.previous_timestep]
            }
        return bundle

    def get_decision(self, results_handle):
        return []



# -*- coding: utf-8 -*-
"""Data access modules for loading system-of-systems model configuration
"""

# import classes for access like ::
#         from smif.data_layer import DataHandle`
from smif.data_layer.data_array import DataArray
from smif.data_layer.data_handle import DataHandle
from smif.data_layer.store import Store

# Define what should be imported as * ::
#         from smif.data_layer import *
__all__ = ['DataArray', 'DataHandle', 'Store']



# -*- coding: utf-8 -*-
"""This module acts as a bridge to the sector models from the controller

The :class:`SectorModel` exposes several key methods for running wrapped
sector models.  To add a sector model to an instance of the framework,
first implement :class:`SectorModel`.

Key Functions
=============
This class performs several key functions which ease the integration of sector
models into the system-of-systems framework.

The user must implement the various abstract functions throughout the class to
provide an interface to the sector model, which can be called upon by the
framework. From the model's perspective, :class:`SectorModel` provides a bridge
from the sector-specific problem representation to the general representation
which allows reasoning across infrastructure systems.

The key functions include:

* converting input/outputs to/from geographies/temporal resolutions
* converting control vectors from the decision layer of the framework, to
  asset Interventions specific to the sector model
* returning scalar/vector values to the framework to enable measurements of
  performance, particularly for the purposes of optimisation and rule-based
  approaches

"""
from abc import ABCMeta, abstractmethod

from smif.model.model import Model

__author__ = "Will Usher, Tom Russell"
__copyright__ = "Will Usher, Tom Russell"
__license__ = "mit"


class SectorModel(Model, metaclass=ABCMeta):
    """A representation of the sector model with inputs and outputs

    Implement this class to enable integration of the wrapped simulation model
    into a system-of-system model.

    Arguments
    ---------
    name : str
        The unique name of the sector model

    Notes
    -----

    Implement the various abstract functions throughout the class to
    provide an interface to the simulation model, which can then be called
    upon by the framework.

    The key methods in the SectorModel class which must be overridden are:

    - :py:meth:`SectorModel.simulate`
    - :py:meth:`SectorModel.extract_obj`

    An implementation may also override:

    - :py:meth:`SectorModel.before_model_run`

    """
    def __init__(self, name):
        super().__init__(name)

    def before_model_run(self, data):
        """Implement this method to conduct pre-model run tasks

        Arguments
        ---------
        data: smif.data_layer.DataHandle
            Access parameter values (before any model is run, no dependency
            input data or state is guaranteed to be available)
            Access decision/system state (i.e. initial_conditions)
        """
        pass

    @abstractmethod
    def simulate(self, data):
        """Implement this method to run the model

        Arguments
        ---------
        data: smif.data_layer.DataHandle
            Access state, parameter values, dependency inputs, results and
            interventions

        Notes
        -----
        See docs on :class:`~smif.data_layer.data_handle.DataHandle` for details of how to
        access inputs, parameters and state and how to set results.

        ``interval``
            should reference an id from the interval set corresponding to
            the output parameter, as specified in model configuration
        ``region``
            should reference a region name from the region set corresponding to
            the output parameter, as specified in model configuration

        To obtain simulation model data in this method,
        use the methods such as::

            parameter_value = data.get_parameter('my_parameter_name')

        Other useful methods are
        :meth:`~smif.data_layer.data_handle.DataHandle.get_base_timestep_data`,
        :meth:`~smif.data_layer.data_handle.DataHandle.get_previous_timestep_data`,
        :meth:`~smif.data_layer.data_handle.DataHandle.get_parameter`,
        :meth:`~smif.data_layer.data_handle.DataHandle.get_data`,
        :meth:`~smif.data_layer.data_handle.DataHandle.get_parameters` and
        :meth:`~smif.data_layer.data_handle.DataHandle.get_results`.

        :meth:`~smif.data_layer.data_handle.DataHandle.get_state` returns a list
        of intervention dict for the current timestep
        :meth:`~smif.data_layer.data_handle.DataHandle.get_current_interventions`
        returns a list of dict where each dict is an intervention

        """



"""Conversion data fixtures
"""
import numpy as np
from pytest import fixture


@fixture(scope='module')
def year_to_month_coefficients():
    """From one year to 12 months

    (apportions)
    """
    return np.array([[31, 28, 31, 30, 31, 31, 30, 30, 31, 31, 30, 31]], dtype=np.float).T / 365


@fixture(scope='module')
def month_to_year_coefficients():
    """From 12 months to one year
    """
    return np.ones((1, 12), dtype=np.float)


@fixture(scope='module')
def month_to_season_coefficients():
    """
    12 months to four seasons (winter is December, January, Feb)

    Sum value for each month into season
    """
    coef = np.array([
        [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],  # winter
        [0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],  # spring
        [0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0],  # summer
        [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0]  # autumn
    ])
    return coef.T


@fixture(scope='module')
def season_to_month_coefficients():
    """
    12 months to four seasons (winter is December, January, Feb)

    To convert from seasons to months, find the proportion of each season that
    corresponds to the relevant month.

    E.g. winter to january is (duration of Jan / total duration of winter)
    """
    coef = np.array(
        # winter
        #     spring
        #        summer
        #           autumn
        [[31, 0, 0, 0],  # January
         [28, 0, 0, 0],  # Feb
         [0, 31, 0, 0],  # March
         [0, 30, 0, 0],  # April
         [0, 31, 0, 0],  # May
         [0, 0, 30, 0],  # June
         [0, 0, 31, 0],  # July
         [0, 0, 31, 0],  # August
         [0, 0, 0, 30],  # September
         [0, 0, 0, 31],  # October
         [0, 0, 0, 30],  # November
         [31, 0, 0, 0]]   # December
    )

    days_in_seasons = np.array([
        31+31+28,  # winter
        31+30+31,  # spring
        30+31+31,  # summer
        30+31+30  # autumn
    ], dtype=float)

    return np.transpose(coef / days_in_seasons)


@fixture(scope='function')
def months():
    data = [
        {'name': 'jan', 'interval': [['P0M', 'P1M']]},
        {'name': 'feb', 'interval': [['P1M', 'P2M']]},
        {'name': 'mar', 'interval': [['P2M', 'P3M']]},
        {'name': 'apr', 'interval': [['P3M', 'P4M']]},
        {'name': 'may', 'interval': [['P4M', 'P5M']]},
        {'name': 'jun', 'interval': [['P5M', 'P6M']]},
        {'name': 'jul', 'interval': [['P6M', 'P7M']]},
        {'name': 'aug', 'interval': [['P7M', 'P8M']]},
        {'name': 'sep', 'interval': [['P8M', 'P9M']]},
        {'name': 'oct', 'interval': [['P9M', 'P10M']]},
        {'name': 'nov', 'interval': [['P10M', 'P11M']]},
        {'name': 'dec', 'interval': [['P11M', 'P12M']]},
    ]
    return data


@fixture
def seasons():
    # NB "winter" is split into two pieces around the year end
    data = [
        {'name': 'winter', 'interval': [['P0M', 'P2M'], ['P11M', 'P12M']]},
        {'name': 'spring', 'interval': [['P2M', 'P5M']]},
        {'name': 'summer', 'interval': [['P5M', 'P8M']]},
        {'name': 'autumn', 'interval': [['P8M', 'P11M']]},
    ]
    return data


@fixture(scope='function')
def twenty_four_hours():
    data = [
        {'name': '1_0', 'interval': [['PT0H', 'PT1H']]},
        {'name': '1_1', 'interval': [['PT1H', 'PT2H']]},
        {'name': '1_2', 'interval': [['PT2H', 'PT3H']]},
        {'name': '1_3', 'interval': [['PT3H', 'PT4H']]},
        {'name': '1_4', 'interval': [['PT4H', 'PT5H']]},
        {'name': '1_5', 'interval': [['PT5H', 'PT6H']]},
        {'name': '1_6', 'interval': [['PT6H', 'PT7H']]},
        {'name': '1_7', 'interval': [['PT7H', 'PT8H']]},
        {'name': '1_8', 'interval': [['PT8H', 'PT9H']]},
        {'name': '1_9', 'interval': [['PT9H', 'PT10H']]},
        {'name': '1_10', 'interval': [['PT10H', 'PT11H']]},
        {'name': '1_11', 'interval': [['PT11H', 'PT12H']]},
        {'name': '1_12', 'interval': [['PT12H', 'PT13H']]},
        {'name': '1_13', 'interval': [['PT13H', 'PT14H']]},
        {'name': '1_14', 'interval': [['PT14H', 'PT15H']]},
        {'name': '1_15', 'interval': [['PT15H', 'PT16H']]},
        {'name': '1_16', 'interval': [['PT16H', 'PT17H']]},
        {'name': '1_17', 'interval': [['PT17H', 'PT18H']]},
        {'name': '1_18', 'interval': [['PT18H', 'PT19H']]},
        {'name': '1_19', 'interval': [['PT19H', 'PT20H']]},
        {'name': '1_20', 'interval': [['PT20H', 'PT21H']]},
        {'name': '1_21', 'interval': [['PT21H', 'PT22H']]},
        {'name': '1_22', 'interval': [['PT22H', 'PT23H']]},
        {'name': '1_23', 'interval': [['PT23H', 'PT24H']]},
    ]
    return data


@fixture(scope='function')
def one_day():
    data = [
        {'name': 'one_day', 'interval': [['P0D', 'P1D']]},
    ]
    return data


@fixture(scope='function')
def one_year():
    data = [
        {'name': 'one_year', 'interval': [['P0Y', 'P1Y']]},
    ]
    return data


@fixture(scope='function')
def monthly_data():
    """[31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    """
    data = np.array([
        31,
        28,
        31,
        30,
        31,
        30,
        31,
        31,
        30,
        31,
        30,
        31
    ])

    return data


@fixture(scope='function')
def monthly_data_as_seasons():
    return np.array([
        31+31+28,
        31+30+31,
        30+31+31,
        30+31+30
    ], dtype=float)


@fixture(scope='function')
def remap_month_data():
    data = np.array([
        31+31+28,  # Dec, Jan, Feb
        31+30+31,  # Mar, Apr, May
        30+31+31,  # Jun, Jul, Aug
        30+31+30  # Sep, Oct, Nov
    ], dtype=float) / 3

    return data


@fixture(scope='function')
def remap_month_data_as_months():
    data = np.array([
        30.666666666,
        29.666666666,
        29.666666666,
        29.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666,
        30.666666666
    ])
    return data


@fixture(scope='function')
def regions_rect():
    """Return single region covering 2x1 area::

        |```````|
        |   0   |
        |.......|

    """
    return [
        {
            'name': 'zero',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'zero'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 2], [1, 2], [1, 0]]]
                }
            }
        }
    ]


@fixture(scope='function')
def regions_half_squares():
    """Return two adjacent square regions::

        |```|```|
        | A | B |
        |...|...|

    """
    return [
        {
            'name': 'a',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
                }
            }
        },
        {
            'name': 'b',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'b'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 1], [0, 2], [1, 2], [1, 1]]]
                }
            }
        }
    ]


@fixture(scope='function')
def regions():
    """Return data structure for test regions/shapes
    """
    return [
        {
            'name': 'unit',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'unit'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
                }
            }
        },
        {
            'name': 'half',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'half'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 0.5], [1, 0.5], [1, 0]]]
                }
            }
        },
        {
            'name': 'two',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'two'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 2], [1, 2], [1, 0]]]
                }
            }
        }
    ]


@fixture(scope='function')
def regions_single_half_square():
    """Return single half-size square region::

        |```|
        | A |
        |...|

    """
    return [
        {
            'name': 'a',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'a'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 1], [1, 1], [1, 0]]]
                }
            }
        }
    ]


@fixture(scope='function')
def regions_half_triangles():
    """Return regions split diagonally::

        |``````/|
        | 0 / 1 |
        |/......|

    """
    return [
        {
            'name': 'zero',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'zero'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 0], [0, 2], [1, 0]]]
                }
            }
        },
        {
            'name': 'one',
            'feature': {
                'type': 'Feature',
                'properties': {'name': 'one'},
                'geometry': {
                    'type': 'Polygon',
                    'coordinates': [[[0, 2], [1, 2], [1, 0]]]
                }
            }
        }
    ]



"""Implements a composite scenario/sector model/system-of-systems model

Begin by declaring the atomic units (scenarios and sector models) which make up
a the composite system-of-systems model and then add these to the composite.
Declare dependencies by using the ``add_dependency()`` method, passing in a
reference to the source model object, and a pointer to the model output
and sink parameter name for the destination model.

Run the model by calling the ``simulate()`` method, passing in a dictionary
containing data for any free hanging model inputs, not linked through a
dependency. A fully defined SosModel should have no hanging model inputs, and
can therefore be called using ``simulate()`` with no arguments.

Responsibility for passing required data to the contained models lies with the
calling class. This means data is only ever passed one layer down.
This simplifies the interface, and allows as little or as much hiding of data,
dependencies and model inputs as required.

Example
-------
A very simple example with just one scenario:

>>> elec_scenario = ScenarioModel('scenario')
>>> elec_scenario.add_output('demand', 'national', 'annual', 'GWh')
>>> sos_model = SosModel('simple')
>>> sos_model.add_model(elec_scenario)
>>> sos_model.simulate(2010)
{'scenario': {'demand': array([123])}}

A more comprehensive example with one scenario and one scenario model:

>>> elec_scenario = ScenarioModel('scenario')
>>> elec_scenario.add_output('demand', 'national', 'annual', 'GWh')
>>> class EnergyModel(SectorModel):
...   def extract_obj(self):
...     pass
...   def simulate(self, timestep, data):
...     return {self.name: {'cost': data['input'] * 2}}
...
>>> energy_model = EnergyModel('model')
>>> energy_model.add_input('input', 'national', 'annual', 'GWh')
>>> energy_model.add_dependency(elec_scenario, 'demand', 'input', lambda x: x)
>>> sos_model = SosModel('sos')
>>> sos_model.add_model(elec_scenario)
>>> sos_model.add_model(energy_model)
>>> sos_model.simulate(2010)
{'model': {'cost': array([[246]])}, 'scenario': {'demand': array([[123]])}}

"""

# import classes for access like ::
#         from smif.model import Model
from smif.model.model import Model, ModelOperation
from smif.model.sector_model import SectorModel
from smif.model.model import ScenarioModel
from smif.model.sos_model import SosModel


# Define what should be imported as * ::
#         from smif.model import *
__all__ = ['Model', 'ModelOperation', 'SectorModel', 'ScenarioModel', 'SosModel']



"""smif runs models over a series of planning timesteps.

:class:`~smif.metadata.timestep.RelativeTimestep` is used to specify one timestep relative to
another in the context of the set of planning timesteps.

For example, if a model needs to know its own outputs from a previous year, we can specify the
self-dependency using a relative timestep.
"""
from enum import Enum

from smif.exception import SmifTimestepResolutionError


class RelativeTimestep(Enum):
    """Specify current, previous or base year timestep
    """
    CURRENT = 'CURRENT'  # current planning timestep
    PREVIOUS = 'PREVIOUS'  # previous planning timestep
    BASE = 'BASE'  # base year planning timestep
    ALL = 'ALL'  # all planning timesteps

    @classmethod
    def from_name(cls, name):
        if name == 'CURRENT':
            return cls.CURRENT
        elif name == 'PREVIOUS':
            return cls.PREVIOUS
        elif name == 'BASE':
            return cls.BASE
        elif name == 'ALL':
            return cls.ALL
        raise ValueError("Relative timestep '%s' is not recognised" % name)

    def resolve_relative_to(self, timestep, timesteps):
        """Resolve a relative timestep with respect to a given timestep and
        sequence of timesteps.

        Parameters
        ----------
        timestep : int
        timesteps : list of int
        """
        if timestep not in timesteps:
            raise SmifTimestepResolutionError(
                "Timestep {} is not present in {}".format(timestep, timesteps))

        # default None, e.g. for ALL
        relative_timestep = None

        if self.name == 'CURRENT':
            relative_timestep = timestep
        elif self.name == 'PREVIOUS':
            try:
                relative_timestep = element_before(timestep, timesteps)
            except ValueError:
                raise SmifTimestepResolutionError(
                    "{} has no previous timestep in {}".format(timestep, timesteps))
        elif self.name == 'BASE':
            relative_timestep = timesteps[0]

        return relative_timestep


def element_before(element, list_):
    """Return the element before a given element in a list, or None if the
    given element is first or not in the list.
    """
    if element not in list_ or element == list_[0]:
        raise ValueError("No element before {} in {}".format(element, list_))
    else:
        index = list_.index(element)
        return list_[index - 1]



"""Test metadata specification
"""
from pytest import fixture, raises
from smif.metadata import Coordinates, Spec


class TestSpec():
    """A Spec serves as metadata for a model input/parameter/output

    Describes an Exchange item (OpenMI)

    dtype - see numpy.dtypes
    Quality (OpenMI) categorical
    Quantity (OpenMI) numeric
    """

    @fixture(scope='function')
    def spec(self):
        spec = Spec(
            name='population',
            description='Population by age class',
            coords=[
                Coordinates('countries', ["England", "Wales"]),
                Coordinates('age', [">30", "<30"])
            ],
            dtype='int',
            abs_range=(0, float('inf')),
            exp_range=(10e6, 10e9),
            unit='people'
        )
        return spec

    def test_construct(self, spec):
        """A Spec has:
        - coords: coordinates that label each point - list of Coordinates, one for each dim
        - name
        - dtype
        - absolute range: (optional) for numerical types, to raise error if exceeded
        - expected range: (optional) for numerical types, to raise warning if exceeded
        - unit
        - acceptable values

        The DataArray it describes may be sparse
        """
        assert spec.name == 'population'
        assert spec.description == 'Population by age class'
        assert spec.unit == 'people'
        assert spec.abs_range == (0, float('inf'))
        assert spec.exp_range == (10e6, 10e9)
        assert spec.dtype == 'int'
        assert spec.shape == (2, 2)
        assert spec.ndim == 2
        assert spec.dims == ['countries', 'age']
        assert spec.coords == [
            Coordinates('countries', ["England", "Wales"]),
            Coordinates('age', [">30", "<30"])
        ]

    def test_rename(self, spec):
        """Allow setting a Spec name
        """
        assert spec.name == 'population'
        spec.name = 'test'
        assert spec.name == 'test'

    def test_dim_coords_method(self, spec):
        assert spec.dim_coords('countries') == Coordinates('countries', ["England", "Wales"])
        with raises(KeyError) as err:
            spec.dim_coords('does not exist')

        assert "Could not find dim 'does not exist' in Spec 'population'" in str(err)

        with raises(TypeError) as err:
            spec.dim_coords(['wrong type'])

        assert "Expected string as argument, instead received <class 'list'>" in str(err)

        spec._dims = ['countries', 'age', 'no coords']
        with raises(KeyError) as err:
            spec.dim_coords('no coords')

        assert "Coords not found for dim 'no coords', in Spec 'population'" in str(err)

    def test_dim_names(self, spec):
        """Names of each coordinate in a given dimension
        """
        assert spec.dim_names('countries') == ["England", "Wales"]

    def test_dim_elements(self, spec):
        """Elements of each coordinate in a given dimension
        """
        assert spec.dim_elements('countries') == [{'name': "England"}, {'name': "Wales"}]

    def test_from_dict(self):
        """classmethod to construct from serialisation
        """
        spec = Spec.from_dict({
            'name': 'population',
            'description': 'Population by age class',
            'dims': ['countries', 'age'],
            'coords': {
                'age': [">30", "<30"],
                'countries': ["England", "Wales"]
            },
            'dtype': 'int',
            'abs_range': (0, float('inf')),
            'exp_range': (10e6, 10e9),
            'unit': 'people'
        })
        assert spec.name == 'population'
        assert spec.description == 'Population by age class'
        assert spec.unit == 'people'
        assert spec.abs_range == (0, float('inf'))
        assert spec.exp_range == (10e6, 10e9)
        assert spec.dtype == 'int'
        assert spec.shape == (2, 2)
        assert spec.ndim == 2
        assert spec.dims == ['countries', 'age']
        assert spec.coords == [
            Coordinates('countries', ["England", "Wales"]),
            Coordinates('age', [">30", "<30"])
        ]

    def test_from_dict_defaults(self):
        """classmethod to construct from serialisation
        """
        spec = Spec.from_dict({
            'dtype': 'int'
        })
        assert spec.name is None
        assert spec.description is None
        assert spec.unit is None
        assert spec.abs_range is None
        assert spec.exp_range is None
        assert spec.dtype == 'int'
        assert spec.shape == ()
        assert spec.ndim == 0
        assert spec.dims == []
        assert spec.coords == []

    def test_to_dict(self):
        actual = Spec(
            name='population',
            description='Population by age class',
            coords=[Coordinates('countries', ["England", "Wales"])],
            dtype='int',
            abs_range=(0, float('inf')),
            exp_range=(10e6, 10e9),
            unit='people'
        ).as_dict()
        expected = {
            'name': 'population',
            'description': 'Population by age class',
            'dims': ['countries'],
            'coords': {'countries': ["England", "Wales"]},
            'dtype': 'int',
            'abs_range': (0, float('inf')),
            'exp_range': (10e6, 10e9),
            'unit': 'people'
        }
        assert actual == expected

    def test_empty_dtype_error(self):
        """A Spec must be constructed with a dtype
        """
        with raises(ValueError) as ex:
            Spec(
                name='test',
                coords=[Coordinates('countries', ["England", "Wales"])]
            )
        assert "dtype must be provided" in str(ex)

    def test_coords_type_error(self):
        """A Spec must be constructed with a list of Coordinates
        """
        with raises(ValueError) as ex:
            Spec(
                name='test',
                dtype='int',
                coords=["England", "Wales"]
            )
        assert "coords may be a dict[str,list] or a list[Coordinates]" in str(ex)

    def test_coords_from_dict(self):
        """A Spec may be constructed with a dict
        """
        spec = Spec(
            name='test',
            dtype='int',
            dims=['countries'],
            coords={'countries': ["England", "Wales"]}
        )
        assert spec.shape == (2,)
        assert spec.coords[0].name == 'countries'
        assert spec.coords[0].ids == ["England", "Wales"]

    def test_coords_from_dict_error(self):
        """A Spec constructed with a dict must have dims
        """
        with raises(ValueError) as ex:
            Spec(
                name='test',
                dtype='int',
                coords={'countries': ["England", "Wales"]}
            )
        assert "dims must be specified" in str(ex)

        with raises(ValueError) as ex:
            Spec(
                name='test',
                dtype='int',
                dims=['countries', 'age'],
                coords={'countries': ["England", "Wales"]}
            )
        assert "dims must match the keys in coords" in str(ex)

        with raises(ValueError) as ex:
            Spec(
                name='test',
                dtype='int',
                dims=['countries'],
                coords={
                    'countries': ["England", "Wales"],
                    'age': [">30", "<30"]
                }
            )
        assert "dims must match the keys in coords" in str(ex)

    def test_duplicate_dims_error(self):
        """A Spec must not have duplicate dimension names
        """
        with raises(ValueError) as ex:
            Spec(
                dtype='int',
                dims=['countries', 'countries'],
                coords={'countries': ["Scotland", "Northern Ireland"]}
            )
        assert "duplicate dims" in str(ex)

        with raises(ValueError) as ex:
            Spec(
                dtype='int',
                coords=[
                    Coordinates('countries', ['Scotland', 'Northern Ireland']),
                    Coordinates('countries', ['Scotland', 'Northern Ireland']),
                ]
            )
        assert "duplicate dims" in str(ex)

    def test_coords_from_list_error(self):
        """A Spec constructed with a dict must have dims
        """
        with raises(ValueError) as ex:
            Spec(
                name='test',
                dtype='int',
                coords=[Coordinates('countries', ["England", "Wales"])],
                dims=['countries']
            )
        assert "dims are derived" in str(ex)

    def test_ranges_must_be_list_like(self):
        """Absolute and expected ranges must be list or tuple
        """
        with raises(TypeError) as ex:
            Spec(
                dtype='int',
                abs_range='string should fail'
            )
        assert "range must be a list or tuple" in str(ex)

        with raises(TypeError) as ex:
            Spec(
                dtype='int',
                exp_range='string should fail'
            )
        assert "range must be a list or tuple" in str(ex)

    def test_ranges_must_be_len_two(self):
        """Absolute and expected ranges must be length two (min and max)
        """
        with raises(ValueError) as ex:
            Spec(
                dtype='int',
                abs_range=[0]
            )
        assert "range must have min and max values only" in str(ex)

        with raises(ValueError) as ex:
            Spec(
                dtype='int',
                exp_range=[0, 1, 2]
            )
        assert "range must have min and max values only" in str(ex)

    def test_ranges_must_be_min_max(self):
        """Absolute and expected ranges must be in order
        """
        with raises(ValueError) as ex:
            Spec(
                dtype='int',
                abs_range=[2, 1]
            )
        assert "min value must be smaller than max value" in str(ex)

        with raises(ValueError) as ex:
            Spec(
                dtype='int',
                exp_range=[2, 1]
            )
        assert "min value must be smaller than max value" in str(ex)

    def test_eq(self):
        """Equality based on equivalent dtype, dims, coords, unit
        """
        a = Spec(
            name='population',
            coords=[Coordinates('countries', ["England", "Wales"])],
            dtype='int',
            unit='people'
        )
        b = Spec(
            name='pop',
            coords=[Coordinates('countries', ["England", "Wales"])],
            dtype='int',
            unit='people'
        )
        c = Spec(
            name='population',
            coords=[Coordinates('countries', ["England", "Scotland", "Wales"])],
            dtype='int',
            unit='people'
        )
        d = Spec(
            name='population',
            coords=[Coordinates('countries', ["England", "Wales"])],
            dtype='float',
            unit='people'
        )
        e = Spec(
            name='population',
            coords=[Coordinates('countries', ["England", "Wales"])],
            dtype='int',
            unit='thousand people'
        )
        assert a == b
        assert a != c
        assert a != d
        assert a != e



"""Adaptor is a subclass of :class:`~smif.model.model.Model`, to be used for converting
data between units or dimensions.

The method to override is `generate_coefficients`, which accepts two
:class:`~smif.metadata.spec.Spec` definitions.
"""
from abc import ABCMeta, abstractmethod

import numpy as np  # type: ignore
from smif.data_layer.data_array import DataArray
from smif.data_layer.data_handle import DataHandle
from smif.exception import SmifDataNotFoundError
from smif.metadata import Spec
from smif.model import Model


class Adaptor(Model, metaclass=ABCMeta):
    """Abstract Adaptor, to convert inputs/outputs between other Models

    Override method `generate_coefficients`, which accepts two
    :class:`~smif.metadata.spec.Spec` definitions.

    """
    def simulate(self, data_handle: DataHandle):
        """Convert from input to output based on matching variable names
        """
        for from_spec in self.inputs.values():
            if from_spec.name in self.outputs:
                to_spec = self.outputs[from_spec.name]
                coefficients = self.get_coefficients(data_handle, from_spec, to_spec)
                data_in = data_handle.get_data(from_spec.name)
                data_out = self.convert(data_in, to_spec, coefficients)
                data_handle.set_results(to_spec.name, data_out)

    def get_coefficients(self,
                         data_handle: DataHandle,
                         from_spec: Spec,
                         to_spec: Spec) -> np.ndarray:
        """Read coefficients, or generate and save if necessary

        Parameters
        ----------
        data_handle : smif.data_layer.data_handle.DataHandle
        from_spec : smif.metadata.spec.Spec
        to_spec : smif.metadata.spec.Spec

        Returns
        -------
        numpy.ndarray
        """
        from_dim, to_dim = self.get_convert_dims(from_spec, to_spec)
        try:
            coefficients = data_handle.read_coefficients(from_dim, to_dim)
        except SmifDataNotFoundError:
            msg = "Generating coefficients for %s to %s"
            self.logger.info(msg, from_dim, to_dim)

            coefficients = self.generate_coefficients(from_spec, to_spec)
            data_handle.write_coefficients(from_dim, to_dim, coefficients)
        return coefficients

    @abstractmethod
    def generate_coefficients(self, from_spec: Spec, to_spec: Spec) -> np.ndarray:
        """Generate coefficients for a pair of :class:`~smif.metadata.spec.Spec` definitions

        Parameters
        ----------
        from_spec : smif.metadata.spec.Spec
        to_spec : smif.metadata.spec.Spec

        Returns
        -------
        numpy.ndarray
        """
        raise NotImplementedError

    def convert(self,
                data_array: DataArray,
                to_spec: Spec,
                coefficients: np.ndarray):
        """Convert a dataset between :class:`~smif.metadata.spec.Spec` definitions

        Parameters
        ----------
        data: smif.data_layer.data_array.DataArray
        to_spec : smif.metadata.spec.Spec
        coefficients : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """
        data = data_array.data
        from_spec = data_array.spec

        self.logger.debug("Converting from %s to %s.", from_spec.name, to_spec.name)

        from_convert_dim, to_convert_dim = self.get_convert_dims(from_spec, to_spec)

        self.logger.debug("Converting from %s:%s to %s:%s", from_spec.name, from_convert_dim,
                          to_spec.name, to_convert_dim)

        axis = from_spec.dims.index(from_convert_dim)

        try:
            converted = self.convert_with_coefficients(data, coefficients, axis)
        except ValueError as ex:
            if coefficients.shape[0] != data.shape[axis]:
                msg = "Coefficients do not match dimension to convert: %s != %s"
                raise ValueError(msg, coefficients.shape[0], data.shape[axis]) from ex
            else:
                raise ex

        self.logger.debug("Converted total from %s to %s", data.sum(), converted.sum())
        return converted

    @staticmethod
    def convert_with_coefficients(data: np.ndarray,
                                  coefficients: np.ndarray,
                                  axis: int):
        """Unchecked conversion, given data, coefficients and axis

        Parameters
        ----------
        data : numpy.ndarray
        coefficients : numpy.ndarray
        axis : integer
            Axis along which to apply conversion coefficients

        Returns
        -------
        numpy.ndarray
        """
        # Effectively a tensor contraction (the generalisation of dot product to multi-
        # dimensional ndarrays, tensors) implemented using the Einstein summation convention,
        # np.einsum, which lets us be explicit which dimensions we sum along.

        # coefficients are 2D, label these 0 and 1
        coefficient_axes = [0, 1]

        # data is nD, label these (2 to n+1) to avoid collisions
        data_axes = list(range(2, 2+data.ndim))
        # except for the axis to convert: label this 0 to match first dim of coefficients
        data_axes[axis] = 0

        # results are also nD, label these (2 to n+1) identically to data_axes
        result_axes = list(range(2, 2+data.ndim))
        # except for the axis to convert: label this 1 to match second dim of coefficients
        result_axes[axis] = 1

        return np.einsum(coefficients, coefficient_axes, data, data_axes, result_axes)

    @staticmethod
    def get_convert_dims(from_spec, to_spec):
        """Get dims for conversion from a pair of :class:`~smif.metadata.spec.Spec`,
        assuming only a single dimension will be converted.

        Parameters
        ----------
        from_spec : smif.metadata.Spec
        to_spec : smif.metadata.Spec

        Returns
        -------
        tuple(str)
        """

        from_convert_dims = set(from_spec.dims) - set(to_spec.dims)
        assert len(from_convert_dims) == 1, "Expected a single dim for conversion"
        from_convert_dim = from_convert_dims.pop()

        to_convert_dims = set(to_spec.dims) - set(from_spec.dims)
        assert len(to_convert_dims) == 1, "Expected a single dim for conversion"
        to_convert_dim = to_convert_dims.pop()

        return from_convert_dim, to_convert_dim



from pytest import raises
from smif.data_layer.model_loader import ModelLoader


def test_path_not_found():
    """Should error if module file is missing at path
    """
    loader = ModelLoader()
    with raises(FileNotFoundError) as ex:
        loader.load({
            'name': 'test',
            'path': '/path/to/model.py',
            'classname': 'WaterSupplySectorModel'
        })
    msg = "Cannot find '/path/to/model.py' for the 'test' model"
    assert msg in str(ex)



"""Register, ResolutionSet abstract classes to contain metadata and generate conversion
coefficients.

NDimensionalRegister is used in :class:`smif.convert.interval.IntervalAdaptor` and
:class:`smif.convert.region.RegionAdaptor`.
"""
import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict, defaultdict
from typing import Dict, List

import numpy as np  # type: ignore


class ResolutionSet(metaclass=ABCMeta):
    """Abstract class which holds the Resolution definitions
    """
    def __init__(self):
        self.name = ''
        self.description = ''
        self._data = []
        self.logger = logging.getLogger(__name__)

    def as_dict(self):
        """Get a serialisable representation of the object
        """
        return {'name': self.name,
                'description': self.description}

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    @property
    def data(self):
        """Resolution set data

        Returns
        -------
        list
        """
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @abstractmethod
    def get_entry_names(self):
        """Get the names of the entries in the ResolutionSet

        Returns
        -------
        set
            The set of names which identify each entry in the ResolutionSet
        """
        raise NotImplementedError

    @abstractmethod
    def intersection(self, bounds):
        """Return the subset of entries intersecting with the bounds
        """
        raise NotImplementedError

    @abstractmethod
    def get_proportion(self, entry_a, entry_b):
        """Calculate the proportion of `entry_a` and `entry_b`

        Arguments
        ---------
        entry_a : string
            Name of an entry in `ResolutionSet`
        entry_b : string
            Name of an entry in `ResolutionSet`

        Returns
        -------
        float
            The proportion of `entry_a` and `entry_b`
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def coverage(self):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_bounds(entry):
        """Implement this helper method to return bounds from an entry in the register

        Arguments
        ---------
        entry
            An entry from a ResolutionSet

        Returns
        -------
        bounds
            The bounds of the entry
        """
        raise NotImplementedError


class LogMixin(object):

    @property
    def logger(self):
        try:
            logger = self._logger
        except AttributeError:
            name = '.'.join([__name__, self.__class__.__name__])
            logger = logging.getLogger(name)
            self._logger = logger
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logger


class Register(LogMixin, metaclass=ABCMeta):
    """Abstract class which holds the ResolutionSets

    Arguments
    ---------
    axis : int, default=None
        The axis over which operations on the data array are performed

    """
    store = None

    def __init__(self, axis=None):
        self.axis = axis

    @property
    @abstractmethod
    def names(self):
        raise NotImplementedError

    @abstractmethod
    def register(self, resolution_set: ResolutionSet):
        raise NotImplementedError

    @abstractmethod
    def get_coefficients(self, source: str, destination: str):
        raise NotImplementedError

    def convert(self, data: np.ndarray, from_set_name: str, to_set_name: str) -> np.ndarray:
        """Convert a list of data points for a given set to another set

        .. deprecated
                Usage superceded by Adaptor.convert

        Parameters
        ----------
        data: numpy.ndarray
        from_set_name: str
        to_set_name: str

        Returns
        -------
        numpy.ndarray

        """
        coefficients = self.get_coefficients(from_set_name, to_set_name)
        converted = Register.convert_with_coefficients(data, coefficients, self.axis)

        self.logger.debug("Converting from %s to %s.", from_set_name, to_set_name)
        self.logger.debug("Converted value from %s to %s", data.sum(), converted.sum())

        return converted

    @staticmethod
    def convert_with_coefficients(data, coefficients: np.ndarray, axis=None) -> np.ndarray:
        """Convert an array of data using given coefficients, along a given axis

        .. deprecated
                Usage superceded by Adaptor.convert

        Parameters
        ----------
        data: numpy.ndarray
        coefficients: numpy.ndarray
        axis: integer, optional

        Returns
        -------
        numpy.ndarray

        """
        if axis is not None:
            data_count = data.shape[axis]
            if coefficients.shape[0] != data_count:
                msg = "Size of coefficient array does not match source " \
                      "resolution set from data matrix: %s != %s"
                raise ValueError(msg, coefficients.shape[axis], data_count)

        if axis == 0:
            converted = np.dot(coefficients.T, data)
        elif axis == 1:
            converted = np.dot(data, coefficients)
        else:
            converted = np.dot(data, coefficients)

        return converted


class NDimensionalRegister(Register):
    """Abstract class which holds N-Dimensional ResolutionSets

    Arguments
    ---------
    axis : int, default=None
        The axis over which operations on the data array are performed

    """
    def __init__(self, axis=None):
        super().__init__(axis)
        self._register = OrderedDict()  # type: Dict[str, ResolutionSet]
        self._conversions = defaultdict(dict)

    def register(self, resolution_set: ResolutionSet):
        """Add a ResolutionSet to the register

        Parameters
        ----------
        resolution_set : :class:`smif.convert.ResolutionSet`

        Raises
        ------
        ValueError
            If a ResolutionSet of the same name already exists in the register
        """
        if resolution_set.name in self._register:
            msg = "A ResolutionSet named {} has already been loaded"
            raise ValueError(msg.format(resolution_set.name))

        self.logger.info("Registering '%s' with %i items",
                         resolution_set.name,
                         len(resolution_set))

        self._register[resolution_set.name] = resolution_set

    @property
    def names(self) -> List[str]:
        """Names of registered region sets

        Returns
        -------
        sets: list[str]
        """
        return list(self._register.keys())

    def get_entry(self, name: str) -> ResolutionSet:
        """Returns the ResolutionSet of `name`

        Arguments
        ---------
        name : str
            The unique identifier of a ResolutionSet in the register

        Returns
        -------
        smif.convert.ResolutionSet

        """
        if name not in self._register:
            msg = "ResolutionSet '{}' not registered"
            raise ValueError(msg.format(name))
        return self._register[name]

    def _write_coefficients(self, source, destination, data: np.ndarray):
        if self.store:
            self.store.write_coefficients(source, destination, data)
        else:
            msg = "Data interface not available to write coefficients"
            self.logger.warning(msg)

    def get_coefficients(self, source: str, destination: str) -> np.ndarray:
        """Get coefficients representing intersection of sets

        Arguments
        ---------
        source : string
            The name of the source set
        destination : string
            The name of the destination set

        Returns
        -------
        numpy.ndarray
        """
        from_set = self.get_entry(source)
        to_set = self.get_entry(destination)

        if from_set.coverage != to_set.coverage:
            log_msg = "Coverage for '%s' is %d and does not match coverage " \
                    "for '%s' which is %d"
            self.logger.warning(log_msg, from_set.name, from_set.coverage,
                                to_set.name, to_set.coverage)

        coefficients = self.generate_coefficients(from_set, to_set)

        return coefficients

    def generate_coefficients(
            self, from_set: ResolutionSet, to_set: ResolutionSet) -> np.ndarray:
        """Generate coefficients for converting between two :class:`ResolutionSet`s

        Coefficients for converting a single dimension will always be 2D, of shape
        (len(from_set), len(to_set)).

        Parameters
        ----------
        from_set : ResolutionSet
        to_set : ResolutionSet

        Returns
        -------
        numpy.ndarray
        """
        coefficients = np.zeros((len(from_set), len(to_set)), dtype=np.float)
        self.logger.debug("Coefficients array is of shape %s for %s to %s",
                          coefficients.shape, from_set.name, to_set.name)

        from_names = from_set.get_entry_names()
        for to_idx, to_entry in enumerate(to_set):
            for from_idx in from_set.intersection(to_entry):
                from_entry = from_set.data[from_idx]
                proportion = from_set.get_proportion(from_idx, to_entry)

                self.logger.debug("%i percent of %s (#%s) is in %s (#%s)",
                                  proportion * 100,
                                  to_entry.name, to_idx,
                                  from_entry.name, from_idx)
                from_idx = from_names.index(from_entry.name)

                coefficients[from_idx, to_idx] = proportion
        self.logger.debug("Generated %s", coefficients)
        return coefficients



"""Test CSV data store
"""
# pylint: disable=redefined-outer-name
import csv
import os
from tempfile import TemporaryDirectory

import numpy as np
from pytest import fixture, mark, raises
from smif.data_layer.data_array import DataArray
from smif.data_layer.file.file_data_store import CSVDataStore
from smif.exception import SmifDataMismatchError, SmifDataNotFoundError
from smif.metadata import Spec


@fixture
def config_handler(setup_folder_structure):
    handler = CSVDataStore(str(setup_folder_structure))
    return handler


@fixture
def get_scenario_data():
    """Return sample scenario_data
    """
    return [
        {
            'population_count': 100,
            'county': 'oxford',
            'season': 'cold_month',
            'timestep': 2017
        },
        {
            'population_count': 150,
            'county': 'oxford',
            'season': 'spring_month',
            'timestep': 2017
        },
        {
            'population_count': 200,
            'county': 'oxford',
            'season': 'hot_month',
            'timestep': 2017
        },
        {
            'population_count': 210,
            'county': 'oxford',
            'season': 'fall_month',
            'timestep': 2017
        },
    ]


@fixture
def get_faulty_scenario_data():
    """Return faulty sample scenario_data
    """
    return [
        {
            'population_count': 100,
            'county': 'oxford',
            'season': 'cold_month',
            'year': 2017
        },
        {
            'population_count': 150,
            'county': 'oxford',
            'season': 'spring_month',
            'year': 2017
        },
        {
            'population_count': 200,
            'county': 'oxford',
            'season': 'hot_month',
            'year': 2017
        },
        {
            'population_count': 210,
            'county': 'oxford',
            'season': 'fall_month',
            'year': 2017
        },
    ]


@fixture(scope='function')
def get_remapped_scenario_data():
    """Return sample scenario_data
    """
    data = [
        {
            'population_count': 100,
            'county': 'oxford',
            'season': 'cold_month',
            'timestep': 2015
        },
        {
            'population_count': 150,
            'county': 'oxford',
            'season': 'spring_month',
            'timestep': 2015
        },
        {
            'population_count': 200,
            'county': 'oxford',
            'season': 'hot_month',
            'timestep': 2015
        },
        {
            'population_count': 210,
            'county': 'oxford',
            'season': 'fall_month',
            'timestep': 2015
        },
        {
            'population_count': 100,
            'county': 'oxford',
            'season': 'cold_month',
            'timestep': 2016
        },
        {
            'population_count': 150,
            'county': 'oxford',
            'season': 'spring_month',
            'timestep': 2016
        },
        {
            'population_count': 200,
            'county': 'oxford',
            'season': 'hot_month',
            'timestep': 2016
        },
        {
            'population_count': 200,
            'county': 'oxford',
            'season': 'fall_month',
            'timestep': 2016
        }
    ]
    return data


@fixture
def scenario_spec():
    return Spec(
        name='population_count',
        unit='people',
        dtype='int',
        dims=['county', 'season'],
        coords={
            'county': ['oxford'],
            'season': ['cold_month', 'spring_month', 'hot_month', 'fall_month']
        }
    )


class TestReadState:
    def test_read_state(self, config_handler):
        handler = config_handler

        modelrun_name = 'modelrun'
        timestep = 2010
        decision_iteration = 0
        dir_ = os.path.join(handler.results_folder, modelrun_name)
        path = os.path.join(dir_, 'state_2010_decision_0.csv')
        os.makedirs(dir_, exist_ok=True)
        with open(path, 'w') as state_fh:
            state_fh.write("build_year,name\n2010,power_station")

        actual = handler.read_state(modelrun_name, timestep, decision_iteration)
        expected = [{'build_year': 2010, 'name': 'power_station'}]
        assert actual == expected


def _write_scenario_csv(base_folder, data, keys):
    key = 'population_high.csv'
    filepath = os.path.join(str(base_folder), 'data', 'scenarios', key)
    with open(filepath, 'w+') as output_file:
        dict_writer = csv.DictWriter(output_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(data)
    return key


class TestScenarios:
    """Scenario data should be readable, metadata is currently editable. May move to make it
    possible to import/edit/write data.
    """
    def test_missing_timestep(self, setup_folder_structure, config_handler,
                              get_faulty_scenario_data, scenario_spec):
        """If a scenario file has incorrect keys, raise a friendly error identifying
        missing keys
        """
        key = _write_scenario_csv(setup_folder_structure, get_faulty_scenario_data,
                                  ('population_count', 'county', 'season', 'year'))

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_scenario_variant_data(key, scenario_spec, 2017)
        msg = "Data for 'population_count' expected a column called 'timestep', instead " + \
              "got data columns ['population_count', 'county', 'season', 'year'] and " + \
              "index names [None]"
        assert msg in str(ex)

    def test_data_column_error(self, setup_folder_structure, config_handler, scenario_spec,
                               get_scenario_data):
        data = []
        for datum in get_scenario_data:
            data.append({
                'population_count': datum['population_count'],
                'region': datum['county'],
                'season': datum['season']
            })
        key = _write_scenario_csv(setup_folder_structure, data,
                                  ('population_count', 'region', 'season'))

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_scenario_variant_data(key, scenario_spec)
        msg = "Data for 'population_count' expected a data column called " + \
              "'population_count' and index names ['county', 'season'], instead got data " + \
              "columns ['population_count', 'region', 'season'] and index names [None]"
        assert msg in str(ex)


    def test_scenario_data_validates(self, setup_folder_structure, config_handler,
                                     get_remapped_scenario_data, scenario_spec):
        """Store performs validation of scenario data against raw interval and region data.

        As such `len(region_names) * len(interval_names)` is not a valid size
        of scenario data under cases where resolution definitions contain
        remapping/resampling info (i.e. multiple hours in a year/regions mapped
        to one name).

        The set of unique region or interval names can be used instead.
        """
        key = _write_scenario_csv(setup_folder_structure, get_remapped_scenario_data,
                                  ('population_count', 'county', 'season', 'timestep'))

        expected_data = np.array([[100, 150, 200, 210]], dtype=float)
        expected = DataArray(scenario_spec, expected_data)

        actual = config_handler.read_scenario_variant_data(key, scenario_spec, 2015)
        assert actual == expected


class TestNarrativeVariantData:
    """Narratives, parameters and interventions should be readable, metadata is editable. May
    move to clarify the distinctions here, and extend to specify strategies and contraints.
    """
    def test_narrative_data(self, setup_folder_structure, config_handler, get_narrative):
        """ Test to dump a narrative (yml) data-file and then read the file
        using the datafile interface. Finally check the data shows up in the
        returned dictionary.
        """
        basefolder = setup_folder_structure
        filepath = os.path.join(
            str(basefolder), 'data', 'narratives', 'central_planning.csv')
        with open(filepath, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['homogeneity_coefficient'])
            writer.writeheader()
            writer.writerow({'homogeneity_coefficient': 8})

        spec = Spec.from_dict({
            'name': 'homogeneity_coefficient',
            'description': "How homegenous the centralisation process is",
            'absolute_range': [0, 1],
            'expected_range': [0, 1],
            'unit': 'percentage',
            'dtype': 'float'
        })

        actual = config_handler.read_narrative_variant_data('central_planning.csv', spec)
        assert actual == DataArray(spec, np.array(8, dtype=float))

    def test_narrative_data_missing(self, config_handler):
        """Should raise a SmifDataNotFoundError if narrative has no data
        """
        with raises(SmifDataNotFoundError):
            config_handler.read_narrative_variant_data('does not exist', None)

    def test_default_data_mismatch(self, config_handler, get_sector_model_parameter_defaults):
        parameter_name = 'smart_meter_savings'
        spec = get_sector_model_parameter_defaults[parameter_name].spec
        path = os.path.join(config_handler.data_folders['parameters'], 'default.csv')
        with open(path, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=[parameter_name])
            writer.writeheader()
            for i in range(4):
                writer.writerow({parameter_name: i})

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_model_parameter_default('default.csv', spec)

        msg = "Data for 'smart_meter_savings' should contain a single value, instead got " + \
              "4 while reading from"
        assert msg in str(ex)

    def test_error_duplicate_rows_single_index(self, config_handler):
        spec = Spec(
            name='test',
            dims=['a'],
            coords={'a': [1, 2]},
            dtype='int'
        )
        path = os.path.join(config_handler.data_folders['parameters'], 'default.csv')
        with open(path, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['test', 'a'])
            writer.writeheader()
            writer.writerows([
                {'a': 1, 'test': 0},
                {'a': 2, 'test': 1},
                {'a': 1, 'test': 2},
            ])

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_model_parameter_default('default.csv', spec)

        msg = "Data for 'test' contains duplicate values at [{'a': 1}]"
        assert msg in str(ex)

    def test_error_duplicate_rows_multi_index(self, config_handler):
        spec = Spec(
            name='test',
            dims=['a', 'b'],
            coords={'a': [1, 2], 'b': [3, 4]},
            dtype='int'
        )
        path = os.path.join(config_handler.data_folders['parameters'], 'default.csv')
        with open(path, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['test', 'a', 'b'])
            writer.writeheader()
            writer.writerows([
                {'a': 1, 'b': 3, 'test': 0},
                {'a': 2, 'b': 3, 'test': 1},
                {'a': 1, 'b': 4, 'test': 2},
                {'a': 2, 'b': 4, 'test': 3},
                {'a': 2, 'b': 4, 'test': 4},
            ])

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_model_parameter_default('default.csv', spec)

        msg = "Data for 'test' contains duplicate values at [{'a': 2, 'b': 4}]"
        msg_alt = "Data for 'test' contains duplicate values at [{'b': 4, 'a': 2}]"
        assert msg in str(ex) or msg_alt in str(ex)

    def test_error_wrong_name(self, config_handler):
        spec = Spec(
            name='test',
            dims=['a', 'b'],
            coords={'a': [1, 2], 'b': [3, 4]},
            dtype='int'
        )
        path = os.path.join(config_handler.data_folders['parameters'], 'default.csv')
        with open(path, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['wrong_name', 'a', 'b'])
            writer.writeheader()
            writer.writerow({
                'a': 1,
                'b': 2,
                'wrong_name': 0
            })

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_model_parameter_default('default.csv', spec)

        msg = "Data for 'test' expected a data column called 'test' and index names " + \
              "['a', 'b'], instead got data columns ['wrong_name'] and index names ['a', 'b']"
        assert msg in str(ex)

    def test_error_not_full(self, config_handler):
        spec = Spec(
            name='test',
            dims=['a', 'b'],
            coords={'a': [1, 2], 'b': [3, 4]},
            dtype='int'
        )
        path = os.path.join(config_handler.data_folders['parameters'], 'default.csv')
        with open(path, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['test', 'a', 'b'])
            writer.writeheader()
            writer.writerow({
                'a': 1,
                'b': 3,
                'test': 0
            })

        with raises(SmifDataMismatchError) as ex:
            config_handler.read_model_parameter_default('default.csv', spec)

        msg = "Data for 'test' had missing values - read 1 but expected 4 in total, from " + \
              "dims of length {a: 2, b: 2}"
        assert msg in str(ex)



class TestResults:
    """Results from intermediate stages of running ModelRuns should be writeable and readable.
    """
    def test_read_results(self, setup_folder_structure, config_handler):
        """Results from .csv in a folder structure which encodes metadata
        in filenames and directory structure.

        On the pattern of:
            <modelrun_name>/
            <model_name>/
            decision_<id>/
                output_<output_name>_
                timestep_<timestep>.csv
        """
        modelrun = 'energy_transport_baseline'
        model = 'energy_demand'
        decision_iteration = 1
        output = 'electricity_demand'
        timestep = 2020
        output_spec = Spec(
            name='electricity_demand',
            unit='MWh',
            dtype='float',
            dims=['region', 'interval'],
            coords={
                'region': ['oxford'],
                'interval': [1]
            }
        )

        expected_data = np.array([[2.0]])
        expected = DataArray(output_spec, expected_data)
        csv_contents = "region,interval,electricity_demand\noxford,1,2.0\n"

        path = os.path.join(
            str(setup_folder_structure),
            "results",
            modelrun,
            model,
            "decision_{}".format(decision_iteration),
            "output_{}_timestep_{}".format(
                output,
                timestep
            )
        )
        os.makedirs(os.path.dirname(path), exist_ok=True)

        with open(path + '.csv', 'w') as fh:
            fh.write(csv_contents)
        actual = config_handler.read_results(
            modelrun, model, output_spec, timestep, decision_iteration)
        assert actual == expected


@mark.skip(reason="Move to test available_results implementation")
class TestWarmStart:
    """If re-running a ModelRun with warm-start specified explicitly, results should be checked
    for existence and left in place.
    """
    def test_prepare_warm_start(self, setup_folder_structure):
        """ Confirm that the warm start copies previous model results
        and reports the correct next timestep
        """

        modelrun = 'energy_transport_baseline'
        model = 'energy_demand'

        # Setup
        basefolder = setup_folder_structure
        current_interface = CSVDataStore(str(basefolder))

        # Create results for a 'previous' modelrun
        previous_results_path = os.path.join(
            str(setup_folder_structure),
            "results", modelrun, model,
            "decision_none"
        )
        os.makedirs(previous_results_path, exist_ok=True)

        path = os.path.join(
            previous_results_path,
            "output_electricity_demand_timestep_2020.csv")
        with open(path, 'w') as fh:
            fh.write("region,interval,value\noxford,1,4.0\n")

        path = os.path.join(
            previous_results_path,
            "output_electricity_demand_timestep_2025.csv")
        with open(path, 'w') as fh:
            fh.write("region,interval,value\noxford,1,6.0\n")

        path = os.path.join(
            previous_results_path,
            "output_electricity_demand_timestep_2030.csv")
        with open(path, 'w') as fh:
            fh.write("region,interval,value\noxford,1,8.0\n")

        # Prepare warm start
        current_timestep = current_interface.prepare_warm_start(modelrun)

        # Confirm that the function reports the correct timestep where the model
        # should continue
        assert current_timestep == 2030

        # Confirm that previous results (excluding the last timestep) exist
        current_results_path = os.path.join(
            str(setup_folder_structure),
            "results", modelrun, model,
            "decision_none"
        )

        warm_start_results = os.listdir(current_results_path)

        assert 'output_electricity_demand_timestep_2020.csv' in warm_start_results
        assert 'output_electricity_demand_timestep_2025.csv' in warm_start_results
        assert 'output_electricity_demand_timestep_2030.csv' not in warm_start_results

    def test_prepare_warm_start_other_local_storage(self, setup_folder_structure):
        """ Confirm that the warm start does not work when previous
        results were saved using a different local storage type
        """

        modelrun = 'energy_transport_baseline'
        model = 'energy_demand'

        # Setup
        basefolder = setup_folder_structure
        current_interface = CSVDataStore(str(basefolder))

        # Create results for a 'previous' modelrun
        previous_results_path = os.path.join(
            str(setup_folder_structure),
            "results",
            modelrun,
            model
        )
        os.makedirs(previous_results_path, exist_ok=True)

        path = os.path.join(
            previous_results_path,
            "output_electricity_demand_timestep_2020.parquet")
        with open(path, 'w') as fh:
            fh.write("region,interval,value\noxford,1,4.0\n")

        path = os.path.join(
            previous_results_path,
            "output_electricity_demand_timestep_2025.parquet")
        with open(path, 'w') as fh:
            fh.write("region,interval,value\noxford,1,6.0\n")

        path = os.path.join(
            previous_results_path,
            "output_electricity_demand_timestep_2030.parquet")
        with open(path, 'w') as fh:
            fh.write("region,interval,value\noxford,1,8.0\n")

        # Prepare warm start
        current_timestep = current_interface.prepare_warm_start(modelrun)

        # Confirm that the function reports the correct timestep where the model
        # should continue
        assert current_timestep is None

    def test_prepare_warm_start_no_previous_results(self, setup_folder_structure):
        """ Confirm that the warm start does not work when no previous
        results were saved
        """

        modelrun = 'energy_transport_baseline'
        model = 'energy_demand'

        # Setup
        basefolder = setup_folder_structure
        current_interface = CSVDataStore(str(basefolder))

        # Create results for a 'previous' modelrun
        previous_results_path = os.path.join(
            str(setup_folder_structure),
            "results",
            modelrun,
            model
        )
        os.makedirs(previous_results_path, exist_ok=True)

        # Prepare warm start
        current_timestep = current_interface.prepare_warm_start(modelrun)

        # Confirm that the function reports the correct timestep where the model
        # should continue
        assert current_timestep is None

        # Confirm that no results were copied
        current_results_path = os.path.join(
            str(setup_folder_structure),
            "results",
            modelrun,
            model
        )
        os.makedirs(current_results_path, exist_ok=True)
        assert len(os.listdir(current_results_path)) == 0

    def test_prepare_warm_start_no_previous_modelrun(self, setup_folder_structure):
        """ Confirm that the warm start does not work when no previous
        modelrun occured
        """

        modelrun = 'energy_transport_baseline'
        model = 'energy_demand'

        # Setup
        basefolder = setup_folder_structure
        current_interface = CSVDataStore(str(basefolder))

        # Prepare warm start
        current_timestep = current_interface.prepare_warm_start(modelrun)

        # Confirm that the function reports the correct timestep where the model
        # should continue
        assert current_timestep is None

        # Confirm that no results were copied
        current_results_path = os.path.join(
            str(setup_folder_structure),
            "results",
            modelrun,
            model
        )
        os.makedirs(current_results_path, exist_ok=True)
        assert len(os.listdir(current_results_path)) == 0


class TestCoefficients:
    """Dimension conversion coefficients should be cached to disk and read if available.
    """
    def test_read_write(self, config_handler):
        data = np.eye(1000)
        handler = config_handler
        handler.write_coefficients('from_dim_name', 'to_dim_name', data)
        actual = handler.read_coefficients('from_dim_name', 'to_dim_name')
        np.testing.assert_equal(actual, data)

    def test_read_raises(self, config_handler):
        handler = config_handler
        with raises(SmifDataNotFoundError):
            handler.read_coefficients('wrong_dim_name', 'to_dim_name')

    def test_dfi_raises_if_folder_missing(self):
        """Ensure we can write files, even if project directory starts empty
        """
        with TemporaryDirectory() as tmpdirname:
            # start with empty project (no data/coefficients subdirectory)
            with raises(SmifDataNotFoundError):
                CSVDataStore(tmpdirname)



"""A metadata store holds shared metadata for smif model scenarios, inputs, parameters and
outputs:
- units
- dimension definitions
"""
from abc import ABCMeta, abstractmethod


class MetadataStore(metaclass=ABCMeta):
    """A MetaDataStore must implement each of the abstract methods defined in this interface
    """
    # region Units
    @abstractmethod
    def read_unit_definitions(self):
        """Reads custom unit definitions

        Returns
        -------
        list[str]
            Pint-compatible unit definitions
        """

    def write_unit_definitions(self, definitions):
        """Reads custom unit definitions

        Parameters
        ----------
        list[str]
            Pint-compatible unit definitions
        """
    # endregion

    # region Dimensions
    @abstractmethod
    def read_dimensions(self, skip_coords=False):
        """Read dimensions

        Parameters
        ----------
        skip_coords : bool, default False
            If True, skip reading dimension elements (names and metadata)

        Returns
        -------
        list[~smif.metadata.coords.Coords]
        """

    @abstractmethod
    def read_dimension(self, dimension_name, skip_coords=False):
        """Return dimension

        Parameters
        ----------
        dimension_name : str
        skip_coords : bool, default False
            If True, skip reading dimension elements (names and metadata)

        Returns
        -------
        ~smif.metadata.coords.Coords
            A dimension definition (including elements)
        """

    @abstractmethod
    def write_dimension(self, dimension):
        """Write dimension to project configuration

        Parameters
        ----------
        dimension : ~smif.metadata.coords.Coords
        """

    @abstractmethod
    def update_dimension(self, dimension_name, dimension):
        """Update dimension

        Parameters
        ----------
        dimension_name : str
        dimension : ~smif.metadata.coords.Coords
        """

    @abstractmethod
    def delete_dimension(self, dimension_name):
        """Delete dimension

        Parameters
        ----------
        dimension_name : str
        """
    # endregion



"""Test SectorModel and SectorModelBuilder
"""
import smif.sample_project.models.water_supply
from pytest import fixture, raises
from smif.metadata import Spec
from smif.model.sector_model import SectorModel
from smif.sample_project.models.water_supply import WaterSupplySectorModel


@fixture(scope='function')
def sector_model_dict():
    path = smif.sample_project.models.water_supply.__file__
    config = {
        "name": "water_supply",
        "description": 'a description',
        "path": path,
        "classname": "WaterSupplySectorModel",
        "inputs": [
            {
                'name': 'raininess',
                'description': None,
                'abs_range': None,
                'exp_range': None,
                'dims': ['LSOA'],
                'coords': {'LSOA': [1, 2, 3]},
                'dtype': 'float',
                'unit': 'milliliter'
            }
        ],
        "outputs": [
            {
                'name': 'cost',
                'description': None,
                'abs_range': None,
                'exp_range': None,
                'dims': ['LSOA'],
                'coords': {'LSOA': [1, 2, 3]},
                'dtype': 'float',
                'unit': 'million GBP'
            },
            {
                'name': 'water',
                'description': None,
                'abs_range': None,
                'exp_range': None,
                'dims': ['LSOA'],
                'coords': {'LSOA': [1, 2, 3]},
                'dtype': 'float',
                'unit': 'megaliter'
            }
        ],
        "parameters": [
            {
                'name': 'assump_diff_floorarea_pp',
                'description': 'Difference in floor area per person',
                'dims': ['national'],
                'coords': {'national': ['GB']},
                'abs_range': (0.5, 2),
                'exp_range': (0.5, 2),
                'dtype': 'float',
                'unit': '%'
            }
        ]
    }
    return config


@fixture(scope='function')
def sector_model():
    model = WaterSupplySectorModel('water_supply')
    model.description = 'a description'
    model.add_input(
        Spec.from_dict({
            'name': 'raininess',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'milliliter'
        })
    )
    model.add_parameter(
        Spec.from_dict({
            'name': 'assump_diff_floorarea_pp',
            'description': 'Difference in floor area per person',
            'dims': ['national'],
            'coords': {'national': ['GB']},
            'abs_range': (0.5, 2),
            'exp_range': (0.5, 2),
            'dtype': 'float',
            'unit': '%'
        })
    )
    model.add_output(
        Spec.from_dict({
            'name': 'cost',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'million GBP'
        })
    )
    model.add_output(
        Spec.from_dict({
            'name': 'water',
            'dims': ['LSOA'],
            'coords': {'LSOA': [1, 2, 3]},
            'dtype': 'float',
            'unit': 'megaliter'
        })
    )
    return model


class EmptySectorModel(SectorModel):
    """Simulate nothing
    """
    def simulate(self, data):
        return data


@fixture(scope='function')
def empty_sector_model():
    return EmptySectorModel('test_model')


class TestSectorModel():
    """A SectorModel has inputs, outputs, and an implementation
    of simulate
    """
    def test_abstract(self):
        with raises(TypeError) as ex:
            SectorModel('cannot_instantiate')
        assert "Can't instantiate abstract class" in str(ex)

    def test_construct(self, sector_model):
        assert sector_model.description == 'a description'
        assert sector_model.inputs == {
            'raininess': Spec.from_dict({
                'name': 'raininess',
                'dims': ['LSOA'],
                'coords': {'LSOA': [1, 2, 3]},
                'dtype': 'float',
                'unit': 'milliliter'
            })
        }

        spec = Spec.from_dict({
                'name': 'assump_diff_floorarea_pp',
                'description': 'Difference in floor area per person',
                'dims': ['national'],
                'coords': {'national': ['GB']},
                'abs_range': (0.5, 2),
                'exp_range': (0.5, 2),
                'dtype': 'float',
                'unit': '%'
            })

        assert sector_model.parameters == {
            'assump_diff_floorarea_pp': spec
        }
        assert sector_model.outputs == {
            'cost': Spec.from_dict({
                'name': 'cost',
                'dims': ['LSOA'],
                'coords': {'LSOA': [1, 2, 3]},
                'dtype': 'float',
                'unit': 'million GBP'
            }),
            'water': Spec.from_dict({
                'name': 'water',
                'dims': ['LSOA'],
                'coords': {'LSOA': [1, 2, 3]},
                'dtype': 'float',
                'unit': 'megaliter'
            })
        }

    def test_from_dict(self, sector_model_dict):
        """Create using classmethod from config
        """
        sector_model = EmptySectorModel.from_dict(sector_model_dict)
        assert sector_model.name == 'water_supply'

    def test_from_dict_no_inputs(self, sector_model_dict):
        """Default sensibly with missing config values
        """
        sector_model_dict['inputs'] = []
        model = EmptySectorModel.from_dict(sector_model_dict)
        assert model.inputs == {}

    def test_as_dict(self, sector_model, sector_model_dict):
        """Serialise back to dict
        """
        actual = sector_model.as_dict()
        # indifferent up to order of inputs/outputs
        actual['inputs'].sort(key=lambda m: m['name'])
        actual['outputs'].sort(key=lambda m: m['name'])
        actual['parameters'].sort(key=lambda m: m['name'])
        assert actual == sector_model_dict

    def test_add_input(self, empty_sector_model):
        """Add an input spec
        """
        spec = Spec(
            name='input_name',
            dims=['dim'],
            coords={'dim': [0]},
            dtype='int'
        )
        empty_sector_model.add_input(spec)
        assert empty_sector_model.inputs == {'input_name': spec}

    def test_add_output(self, empty_sector_model):
        """Add an output spec
        """
        spec = Spec(
            name='output_name',
            dims=['dim'],
            coords={'dim': [0]},
            dtype='int'
        )
        empty_sector_model.add_output(spec)
        assert empty_sector_model.outputs == {'output_name': spec}

    def test_add_parameter(self, empty_sector_model):
        """Adding a parameter adds a reference to the parameter list entry to
        the model that contains it.
        """
        spec = Spec.from_dict({
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'abs_range': (0, 100),
            'exp_range': (3, 10),
            'dims': ['national'],
            'coords': {'national': ['GB']},
            'dtype': 'float',
            'unit': '%'
        })
        empty_sector_model.add_parameter(spec)
        expected = spec
        actual = empty_sector_model.parameters['smart_meter_savings']
        assert actual == expected

    def test_simulate_exists(self, empty_sector_model):
        """Call simulate
        """
        empty_sector_model.simulate(None)

    def test_before_model_run_exists(self, empty_sector_model):
        """Call before_model_run
        """
        empty_sector_model.before_model_run(None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Holds fixtures for the smif package tests
"""
from __future__ import absolute_import, division, print_function

import json
import logging
import os
from copy import deepcopy

import numpy as np
import pandas as pd
from pytest import fixture
from smif.data_layer import Store
from smif.data_layer.data_array import DataArray
from smif.data_layer.file.file_config_store import _write_yaml_file as dump
from smif.data_layer.memory_interface import (MemoryConfigStore,
                                              MemoryDataStore,
                                              MemoryMetadataStore)
from smif.metadata import Spec

logging.basicConfig(filename='test_logs.log',
                    level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s: %(levelname)-8s %(message)s',
                    filemode='w')


@fixture
def empty_store():
    """Store fixture
    """
    # implement each part using the memory classes, simpler than mocking
    # each other implementation of a part is tested fully by e.g. test_config_store.py
    return Store(
        config_store=MemoryConfigStore(),
        metadata_store=MemoryMetadataStore(),
        data_store=MemoryDataStore()
    )


@fixture
def setup_empty_folder_structure(tmpdir_factory):

    folder_list = ['models', 'results', 'config', 'data']

    config_folders = [
        'dimensions',
        'model_runs',
        'scenarios',
        'sector_models',
        'sos_models',
    ]
    for folder in config_folders:
        folder_list.append(os.path.join('config', folder))

    data_folders = [
        'coefficients',
        'dimensions',
        'initial_conditions',
        'interventions',
        'narratives',
        'scenarios',
        'strategies',
        'parameters'
    ]
    for folder in data_folders:
        folder_list.append(os.path.join('data', folder))

    test_folder = tmpdir_factory.mktemp("smif")

    for folder in folder_list:
        test_folder.mkdir(folder)

    return test_folder


@fixture
def setup_folder_structure(setup_empty_folder_structure, oxford_region, remap_months,
                           initial_system, planned_interventions):
    """

    Returns
    -------
    :class:`LocalPath`
        Path to the temporary folder
    """
    test_folder = setup_empty_folder_structure

    region_file = test_folder.join('data', 'dimensions', 'test_region.geojson')
    region_file.write(json.dumps(oxford_region))

    intervals_file = test_folder.join('data', 'dimensions', 'annual.yml')
    intervals_file.write("""\
- name: '1'
  interval: [[P0Y, P1Y]]
""")

    intervals_file = test_folder.join('data', 'dimensions', 'hourly.yml')
    intervals_file.write("""\
- name: '1'
  interval: [[PT0H, PT1H]]
""")

    initial_conditions_dir = str(test_folder.join('data', 'initial_conditions'))
    dump(initial_conditions_dir, 'init_system', initial_system)

    interventions_dir = str(test_folder.join('data', 'interventions'))
    dump(interventions_dir, 'planned_interventions', planned_interventions)

    dimensions_dir = str(test_folder.join('data', 'dimensions'))
    dump(dimensions_dir, 'remap', remap_months)

    units_file = test_folder.join('data', 'user_units.txt')
    with units_file.open(mode='w') as units_fh:
        units_fh.write("blobbiness = m^3 * 10^6\n")
        units_fh.write("people = [people]\n")
        units_fh.write("mcm = 10^6 * m^3\n")
        units_fh.write("GBP=[currency]\n")

    return test_folder


@fixture
def initial_system():
    """Initial system (interventions with build_date)
    """
    return [
        {'name': 'water_asset_a', 'build_year': 2017},
        {'name': 'water_asset_b', 'build_year': 2017},
        {'name': 'water_asset_c', 'build_year': 2017},
    ]


@fixture
def parameters():
    return [
        {
            'name': 'smart_meter_savings',
            'description': 'The savings from smart meters',
            'absolute_range': (0, 100),
            'suggested_range': (3, 10),
            'default_value': 3,
            'unit': '%'
        }
    ]


@fixture
def planned_interventions():
    """Return pre-specified planning intervention data
    """
    return [
        {
            'name': 'water_asset_a',
            'capacity': {'value': 6, 'unit': 'Ml'},
            'description': 'Existing water treatment plants',
            'location': {'lat': 51.74556, 'lon': -1.240528}
        },
        {
            'name': 'water_asset_b',
            'capacity':  {'value': 6, 'unit': 'Ml'},
            'description': 'Existing water treatment plants',
            'location': {'lat': 51.74556, 'lon': -1.240528}
        },
        {
            'name': 'water_asset_c',
            'capacity': {'value': 6, 'unit': 'Ml'},
            'description': 'Existing water treatment plants',
            'location': {'lat': 51.74556, 'lon': -1.240528}
        },
    ]


@fixture
def oxford_region():
    data = {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {
                "name": "urn:ogc:def:crs:EPSG::27700"
            }
        },
        "features": [
            {
                "type": "Feature",
                "properties": {
                    "name": "oxford"
                },
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [448180, 209366],
                            [449500, 211092],
                            [450537, 211029],
                            [450873, 210673],
                            [451250, 210793],
                            [451642, 210023],
                            [453855, 208466],
                            [454585, 208468],
                            [456077, 207967],
                            [456146, 207738],
                            [456668, 207779],
                            [456708, 207444],
                            [456278, 207122],
                            [456149, 206615],
                            [455707, 206798],
                            [455749, 204521],
                            [456773, 204488],
                            [457014, 204184],
                            [456031, 203475],
                            [456444, 202854],
                            [456087, 202044],
                            [455369, 201799],
                            [454396, 202203],
                            [453843, 201634],
                            [452499, 203209],
                            [452052, 203566],
                            [451653, 203513],
                            [450645, 205137],
                            [449497, 205548],
                            [449051, 206042],
                            [448141, 208446],
                            [448180, 209366]
                        ]
                    ]
                }
            },
        ]
    }
    return data


@fixture
def initial_conditions():
    return [{'name': 'solar_installation', 'build_year': 2017}]


@fixture
def interventions():
    return {
        'solar_installation': {
            'name': 'solar_installation',
            'capacity': 5,
            'capacity_units': 'MW'
        },
        'wind_installation': {
            'name': 'wind_installation',
            'capacity': 4,
            'capacity_units': 'MW'
        }
    }


@fixture
def water_interventions_abc():
    return [
        {
            "name": "water_asset_a",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 1000
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        },
        {
            "name": "water_asset_b",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 1500
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        },
        {
            "name": "water_asset_c",
            "location": "oxford",
            "capital_cost": {
                "units": "£",
                "value": 3000
            },
            "economic_lifetime": {
                "units": "years",
                "value": 25
            },
            "operational_lifetime": {
                "units": "years",
                "value": 25
            }
        }
    ]


@fixture
def model_run():
    """Return sample model_run
    """
    return {
        'name': 'unique_model_run_name',
        'description': 'a description of what the model run contains',
        'stamp': '2017-09-20T12:53:23+00:00',
        'timesteps': [
            2015,
            2020,
            2025
        ],
        'sos_model': 'energy',
        'scenarios': {
            'population': 'High Population (ONS)'
        },
        'strategies': [
            {
                'type': 'pre-specified-planning',
                'name': 'energy_supply',
                'description': 'description of the strategy',
                'model_name': 'energy_supply',
            }
        ],
        'narratives': {
            'technology': [
                'Energy Demand - High Tech'
            ],
            'governance': [
                'Central Planning'
            ]
        }
    }


@fixture
def get_sos_model(sample_narratives):
    """Return sample sos_model
    """
    return {
        'name': 'energy',
        'description': "A system of systems model which encapsulates "
                       "the future supply and demand of energy for the UK",
        'scenarios': [
            'population'
        ],
        'narratives': sample_narratives,
        'sector_models': [
            'energy_demand',
            'energy_supply'
        ],
        'scenario_dependencies': [
            {
                'source': 'population',
                'source_output': 'population_count',
                'sink': 'energy_demand',
                'sink_input': 'population'
            }
        ],
        'model_dependencies': [
            {
                'source': 'energy_demand',
                'source_output': 'gas_demand',
                'sink': 'energy_supply',
                'sink_input': 'natural_gas_demand'
            }
        ]
    }


@fixture
def get_sector_model(annual, hourly, lad):
    """Return sample sector_model
    """
    return {
        'name': 'energy_demand',
        'description': "Computes the energy demand of the"
                       "UK population for each timestep",
        'classname': 'EnergyDemandWrapper',
        'path': '../../models/energy_demand/run.py',
        'inputs': [
            {
                'name': 'population',
                'dtype': 'int',
                'dims': ['lad', 'annual'],
                'coords': {
                    'lad': lad,
                    'annual': annual
                },
                'absolute_range': [0, int(1e12)],
                'expected_range': [0, 100000],
                'unit': 'people'
            }
        ],
        'outputs': [
            {
                'name': 'gas_demand',
                'dtype': 'float',
                'dims': ['lad', 'hourly'],
                'coords': {
                    'lad': lad,
                    'hourly': hourly
                },
                'absolute_range': [0, float('inf')],
                'expected_range': [0.01, 10],
                'unit': 'GWh'
            }
        ],
        'parameters': [
            {
                'name': 'smart_meter_savings',
                'description': "Difference in floor area per person"
                               "in end year compared to base year",
                'absolute_range': [0, float('inf')],
                'expected_range': [0.5, 2],
                'unit': '%',
                'dtype': 'float'
            },
            {
                'name': 'homogeneity_coefficient',
                'description': "How homegenous the centralisation"
                               "process is",
                'absolute_range': [0, 1],
                'expected_range': [0, 1],
                'unit': 'percentage',
                'dtype': 'float'
            }
        ],
        'interventions': [],
        'initial_conditions': []
    }


@fixture
def energy_supply_sector_model(hourly):
    """Return sample sector_model
    """
    return {
        'name': 'energy_supply',
        'description': "Supply system model",
        'classname': 'EnergySupplyWrapper',
        'path': '../../models/energy_supply/run.py',
        'inputs': [
            {
                'name': 'natural_gas_demand',
                'dims': ['lad', 'hourly'],
                'coords': {
                    'lad': ['a', 'b'],
                    'hourly': hourly
                },
                'absolute_range': [0, float('inf')],
                'expected_range': [0, 100],
                'dtype': 'float',
                'unit': 'GWh'
            }
        ],
        'outputs': [],
        'parameters': [],
        'interventions': [],
        'initial_conditions': []
    }


@fixture
def water_supply_sector_model(hourly):
    """Return sample sector_model
    """
    return {
        'name': 'water_supply',
        'description': "Supply system model",
        'classname': 'WaterSupplyWrapper',
        'path': '../../models/water_supply/run.py',
        'inputs': [],
        'outputs': [],
        'parameters': [
            {
                'name': 'clever_water_meter_savings',
                'description': "",
                'absolute_range': [0, 1],
                'expected_range': [0, 0.2],
                'unit': 'percentage',
                'dtype': 'float'
            },
            {
                'name': 'per_capita_water_demand',
                'description': "",
                'absolute_range': [0, float('inf')],
                'expected_range': [0, 0.05],
                'unit': 'Ml/day',
                'dtype': 'float'
            }
        ],
        'interventions': [],
        'initial_conditions': []
    }


@fixture
def get_sector_model_parameter_defaults(get_sector_model):
    """DataArray for each parameter default
    """
    data = {
        'smart_meter_savings': np.array(0.5),
        'homogeneity_coefficient': np.array(0.1)
    }
    for param in get_sector_model['parameters']:
        nda = data[param['name']]
        spec = Spec.from_dict(param)
        data[param['name']] = DataArray(spec, nda)
    return data


@fixture
def get_multidimensional_param():
    spec = Spec.from_dict({
        'name': 'ss_t_base_heating',
        'description': 'Industrial base temperature',
        'default': '../energy_demand/parameters/ss_t_base_heating.csv',
        'unit': '',
        'dims': ['interpolation_params', 'end_yr'],
        'coords': {
            'interpolation_params': ['diffusion_choice', 'value_ey'],
            'end_yr': [2030, 2050]
        },
        'dtype': 'float'
    })
    dataframe = pd.DataFrame([
        {
            'interpolation_params': 'diffusion_choice',
            'end_yr': 2030,
            'ss_t_base_heating': 0
        },
        {
            'interpolation_params': 'diffusion_choice',
            'end_yr': 2050,
            'ss_t_base_heating': 0
        },
        {
            'interpolation_params': 'value_ey',
            'end_yr': 2030,
            'ss_t_base_heating': 15.5
        },
        {
            'interpolation_params': 'value_ey',
            'end_yr': 2050,
            'ss_t_base_heating': 15.5
        },
    ]).set_index(['interpolation_params', 'end_yr'])
    return DataArray.from_df(spec, dataframe)


@fixture
def get_sector_model_no_coords(get_sector_model):
    model = deepcopy(get_sector_model)
    for spec_group in ('inputs', 'outputs', 'parameters'):
        for spec in model[spec_group]:
            try:
                del spec['coords']
            except KeyError:
                pass
    return model


@fixture
def sample_scenarios():
    """Return sample scenario
    """
    return [
        {
            'name': 'population',
            'description': 'The annual change in UK population',
            'provides': [
                {
                    'name': "population_count",
                    'description': "The count of population",
                    'unit': 'people',
                    'dtype': 'int',
                    'dims': ['lad', 'annual']
                },
            ],
            'variants': [
                {
                    'name': 'High Population (ONS)',
                    'description': 'The High ONS Forecast for UK population out to 2050',
                    'data': {
                        'population_count': 'population_high.csv'
                    }
                },
                {
                    'name': 'Low Population (ONS)',
                    'description': 'The Low ONS Forecast for UK population out to 2050',
                    'data': {
                        'population_count': 'population_low.csv'
                    }
                },
            ],
        },
    ]


@fixture
def sample_scenario_data(scenario, get_sector_model, energy_supply_sector_model,
                         water_supply_sector_model):
    scenario_data = {}

    for scenario in [scenario]:
        for variant in scenario['variants']:
            for data_key, data_value in variant['data'].items():
                spec = Spec.from_dict(
                    [provides for provides in scenario['provides']
                     if provides['name'] == data_key][0])
                nda = np.random.random(spec.shape)
                da = DataArray(spec, nda)
                key = (scenario['name'], variant['name'], data_key)
                scenario_data[key] = da

    return scenario_data


@fixture
def get_scenario():
    """Return sample scenario
    """
    return {
        "name": "Economy",
        "description": "Economic projections for the UK",
        "provides": [
            {
                'name': "gva",
                'description': "GVA",
                'dtype': "float",
                'unit': "million GBP"
            }
        ],
        "variants": [
            {
                "name": "Central Economy (High)",
                "data": {
                    "gva": 3,
                }
            }
        ]
    }


@fixture(scope='function')
def get_narrative():
    """Return sample narrative
    """
    return {
        'name': 'technology',
        'description': 'Describes the evolution of technology',
        'provides': {
            'energy_demand': ['smart_meter_savings']
        },
        'variants': [
            {
                'name': 'high_tech_dsm',
                'description': 'High takeup of smart technology on the demand side',
                'data': {
                    'smart_meter_savings': 'high_tech_dsm.csv'
                }
            }
        ]
    }


@fixture
def sample_narratives(get_narrative):
    """Return sample narratives
    """
    return [
        get_narrative,
        {
            'name': 'governance',
            'description': 'Defines the nature of governance and influence upon decisions',
            'provides': {
                'energy_demand': ['homogeneity_coefficient']
            },
            'variants': [
                {
                    'name': 'Central Planning',
                    'description': 'Stronger role for central government in planning and ' +
                                   'regulation, less emphasis on market-based solutions',
                    'data': {
                        'homogeneity_coefficient': 'homogeneity_coefficient.csv'
                    }
                },
            ],
        },
    ]


@fixture
def sample_narrative_data(sample_narratives, get_sector_model, energy_supply_sector_model,
                          water_supply_sector_model):
    narrative_data = {}
    sos_model_name = 'energy'
    sector_models = {}
    sector_models[get_sector_model['name']] = get_sector_model
    sector_models[energy_supply_sector_model['name']] = energy_supply_sector_model
    sector_models[water_supply_sector_model['name']] = water_supply_sector_model

    for narrative in sample_narratives:
        for sector_model_name, param_names in narrative['provides'].items():
            sector_model = sector_models[sector_model_name]
            for param_name in param_names:
                param = _pick_from_list(sector_model['parameters'], param_name)
                for variant in narrative['variants']:
                    spec = Spec.from_dict(param)
                    nda = np.random.random(spec.shape)
                    da = DataArray(spec, nda)
                    key = (sos_model_name, narrative['name'], variant['name'], param_name)
                    narrative_data[key] = da
    return narrative_data


@fixture
def sample_results():
    spec = Spec(name='energy_use', dtype='float')
    data = np.array(1, dtype=float)
    return DataArray(spec, data)


def _pick_from_list(list_, name):
    for item in list_:
        if item['name'] == name:
            return item
    assert False, '{} not found in {}'.format(name, list_)


@fixture
def sample_dimensions(remap_months, hourly, annual, lad):
    """Return sample dimensions
    """
    return [
        {
            'name': 'lad',
            'description': 'Local authority districts for the UK',
            'elements': lad
        },
        {
            'name': 'hourly',
            'description': 'The 8760 hours in the year named by hour',
            'elements': hourly
        },
        {
            'name': 'annual',
            'description': 'One annual timestep, used for aggregate yearly data',
            'elements': annual,
        },
        {
            'name': 'remap_months',
            'description': 'Remapped months to four representative months',
            'elements': remap_months,
        },
        {
            'name': 'technology_type',
            'description': 'Technology dimension for narrative fixture',
            'elements': [
                {'name': 'water_meter'},
                {'name': 'electricity_meter'},
            ]
        },
        {
            'name': 'county',
            'elements': [
                {'name': 'oxford'}
            ]
        },
        {
            'name': 'season',
            'elements': [
                {'name': 'cold_month'},
                {'name': 'spring_month'},
                {'name': 'hot_month'},
                {'name': 'fall_month'}
            ]
        }
    ]


@fixture
def get_dimension():
    return {
        "name": "annual",
        "description": "Single annual interval of 8760 hours",
        "elements":
            [
                {
                    "id": 1,
                    "interval": [["PT0H", "PT8760H"]]
                }
            ]
    }


@fixture
def lad():
    return [{'name': 'a'}, {'name': 'b'}]


@fixture
def hourly():
    return [
        {
            'name': n,
            'interval': [['PT{}H'.format(n), 'PT{}H'.format(n+1)]]
        }
        for n in range(8)  # should be 8760
    ]


@fixture
def annual():
    return [
        {
            'name': 1,
            'interval': [['PT0H', 'PT8760H']]
        }
    ]


@fixture
def remap_months():
    """Remapping four representative months to months across the year

    In this case we have a model which represents the seasons through
    the year using one month for each season. We then map the four
    model seasons 1, 2, 3 & 4 onto the months throughout the year that
    they represent.

    The data will be presented to the model using the four time intervals,
    1, 2, 3 & 4. When converting to hours, the data will be replicated over
    the year.  When converting from hours to the model time intervals,
    data will be averaged and aggregated.

    """
    data = [
        {'name': 'cold_month', 'interval': [['P0M', 'P1M'], ['P1M', 'P2M'], ['P11M', 'P12M']]},
        {'name': 'spring_month', 'interval': [['P2M', 'P3M'], ['P3M', 'P4M'], ['P4M', 'P5M']]},
        {'name': 'hot_month', 'interval': [['P5M', 'P6M'], ['P6M', 'P7M'], ['P7M', 'P8M']]},
        {'name': 'fall_month', 'interval': [['P8M', 'P9M'], ['P9M', 'P10M'], ['P10M', 'P11M']]}
    ]
    return data


@fixture
def minimal_model_run():
    return {
        'name': 'test_modelrun',
        'timesteps': [2010, 2015, 2010],
        'sos_model': 'energy'
    }


@fixture
def strategies():
    return [
        {
            'type': 'pre-specified-planning',
            'description': 'a description',
            'model_name': 'test_model',
            'interventions': [
                {'name': 'a', 'build_year': 2020},
                {'name': 'b', 'build_year': 2025},
            ]
        },
        {
            'type': 'rule-based',
            'description': 'reduce emissions',
            'path': 'planning/energyagent.py',
            'classname': 'EnergyAgent'
        }
    ]


@fixture
def unit_definitions():
    return ['kg = kilograms']


@fixture
def dimension():
    return {'name': 'category', 'elements': [{'name': 1}, {'name': 2}, {'name': 3}]}


@fixture
def conversion_source_spec():
    return Spec(name='a', dtype='float', unit='ml')


@fixture
def conversion_sink_spec():
    return Spec(name='b', dtype='float', unit='ml')


@fixture
def conversion_coefficients():
    return np.array([[1]])


@fixture
def scenario(sample_dimensions):

    return deepcopy({
        'name': 'mortality',
        'description': 'The annual mortality rate in UK population',
        'provides': [
            {
                'name': 'mortality',
                'dims': ['lad'],
                'coords': {'lad': sample_dimensions[0]['elements']},
                'dtype': 'float',
            }
        ],
        'variants': [
            {
                'name': 'low',
                'description': 'Mortality (Low)',
                'data': {
                    'mortality': 'mortality_low.csv',
                },
            }
        ]
    })


@fixture
def scenario_no_coords(scenario):
    scenario = deepcopy(scenario)
    for spec in scenario['provides']:
        try:
            del spec['coords']
        except KeyError:
            pass
    return scenario


@fixture
def narrative_no_coords(get_narrative):
    get_narrative = deepcopy(get_narrative)
    for spec in get_narrative['provides']:
        try:
            del spec['coords']
        except KeyError:
            pass
    return get_narrative


@fixture
def state():
    return [
        {
            'name': 'test_intervention',
            'build_year': 1900
        }
    ]



"""A data store holds the bulk of model setup, intermediate and output data:
- scenario variant and narrative data (including parameter defaults)
- model interventions, initial conditions and state
- conversion coefficients
- results
"""
from abc import ABCMeta, abstractmethod


class DataStore(metaclass=ABCMeta):
    """A DataStore must implement each of the abstract methods defined in this interface
    """
    # region DataArray
    @abstractmethod
    def read_scenario_variant_data(self, key, spec, timestep=None):
        """Read data array

        Parameters
        ----------
        key : str
        spec : ~smif.metadata.spec.Spec
        timestep : int (optional)
            If None, read data for all timesteps

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_scenario_variant_data(self, key, data_array, timestep=None):
        """Write data array

        Parameters
        ----------
        key : str
        data_array : ~smif.data_layer.data_array.DataArray
        timestep : int (optional)
            If None, write data for all timesteps
        """

    @abstractmethod
    def read_narrative_variant_data(self, key, spec, timestep=None):
        """Read data array

        Parameters
        ----------
        key : str
        spec : ~smif.metadata.spec.Spec
        timestep : int (optional)
            If None, read data for all timesteps

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_narrative_variant_data(self, key, data_array, timestep=None):
        """Write data array

        Parameters
        ----------
        key : str
        data_array : ~smif.data_layer.data_array.DataArray
        timestep : int (optional)
            If None, write data for all timesteps
        """

    @abstractmethod
    def read_model_parameter_default(self, key, spec):
        """Read data array

        Parameters
        ----------
        key : str
        spec : ~smif.metadata.spec.Spec

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_model_parameter_default(self, key, data_array):
        """Read data array

        Parameters
        ----------
        key : str
        data_array : ~smif.data_layer.data_array.DataArray

        Returns
        -------
        data_array : ~smif.data_layer.data_array.DataArray
        """
    # endregion

    # region Interventions
    @abstractmethod
    def read_interventions(self, key):
        """Read interventions data for `key`

        Parameters
        ----------
        key : str

        Returns
        -------
        dict[str, dict]
            A dict of intervention dictionaries containing intervention
            attributes keyed by intervention name
        """

    @abstractmethod
    def write_interventions(self, key, interventions):
        """Write interventions data for `key`

        Parameters
        ----------
        key : str
        interventions : dict[str, dict]
        """

    @abstractmethod
    def read_initial_conditions(self, key):
        """Read historical interventions for `key`

        Parameters
        ----------
        key : str

        Returns
        -------
        list[dict]
        """

    @abstractmethod
    def write_initial_conditions(self, key, initial_conditions):
        """Write historical interventions for `key`

        Parameters
        ----------
        key : str
        initial_conditions: list[dict]
        """
    # endregion

    # region State
    @abstractmethod
    def read_state(self, modelrun_name, timestep, decision_iteration=None):
        """Read list of (name, build_year) for a given model_run, timestep,
        decision

        Parameters
        ----------
        model_run_name : str
        timestep : int
        decision_iteration : int, optional

        Returns
        -------
        list[dict]
        """

    @abstractmethod
    def write_state(self, state, modelrun_name, timestep, decision_iteration=None):
        """State is a list of decisions with name and build_year.

        State is output from the DecisionManager

        Parameters
        ----------
        state : list[dict]
        model_run_name : str
        timestep : int
        decision_iteration : int, optional
        """
    # endregion

    # region Conversion coefficients
    @abstractmethod
    def read_coefficients(self, source_dim, destination_dim):
        """Reads coefficients from the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `write_coefficients` implement caching of conversion
        coefficients between a single pair of dimensions.

        Parameters
        ----------
        source_dim : str
            dimension name
        destination_dim : str
            dimension name

        Returns
        -------
        numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """

    @abstractmethod
    def write_coefficients(self, source_dim, destination_dim, data):
        """Writes coefficients to the store

        Coefficients are uniquely identified by their source/destination dimensions.
        This method and `read_coefficients` implement caching of conversion
        coefficients between a single pair of dimensions.

        Parameters
        ----------
        source_dim : str
            dimension name
        destination_dim : str
            dimension name
        data : numpy.ndarray

        Notes
        -----
        To be called from :class:`~smif.convert.adaptor.Adaptor` implementations.
        """
    # endregion

    # region Results
    @abstractmethod
    def read_results(self, modelrun_name, model_name, output_spec, timestep=None,
                     decision_iteration=None):
        """Return results of a model from a model_run for a given output at a timestep and
        decision iteration

        Parameters
        ----------
        model_run_id : str
        model_name : str
        output_spec : ~smif.metadata.spec.Spec
        timestep : int, default=None
        decision_iteration : int, default=None

        Returns
        -------
        ~smif.data_layer.data_array.DataArray
        """

    @abstractmethod
    def write_results(self, data, modelrun_name, model_name, timestep=None,
                      decision_iteration=None):
        """Write results of a `model_name` in `model_run_name` for a given `output_name`

        Parameters
        ----------
        data_array : ~smif.data_layer.data_array.DataArray
        model_run_id : str
        model_name : str
        timestep : int, optional
        decision_iteration : int, optional
        """

    @abstractmethod
    def available_results(self, modelrun_name):
        """List available results from a model run

        Returns
        -------
        list[tuple]
             Each tuple is (timestep, decision_iteration, model_name, output_name)
        """
    # endregion



"""Metadata describes the details of data structures to be exchanged between models.

- :class:`~smif.metadata.spec.Spec` provides metadata to describe multi-dimensional variables,
  with an API modelled on `xarray <http://xarray.pydata.org/en/stable/>`_
- :class:`~smif.metadata.coordinates.Coordinates` label each dimension in a
  :class:`~smif.metadata.spec.Spec`. The elements of a set of
  :class:`~smif.metadata.coordinates.Coordinates` correspond to ElementSets under the
  `OGC® Open Modelling Interface (OpenMI) Interface Standard
  <http://www.opengeospatial.org/standards/openmi>`_
"""

# import classes here if they should be accessed at the subpackage level, for example ::
#         from smif.metadata import Spec
from smif.metadata.coordinates import Coordinates
from smif.metadata.timestep import RelativeTimestep
from smif.metadata.spec import Spec

# Define what should be imported as * ::
#         from smif.metadata import *
__all__ = ['Coordinates', 'RelativeTimestep', 'Spec']



"""Model abstract class
"""
import sys
from abc import ABCMeta, abstractmethod
from enum import Enum
from logging import getLogger

from smif.metadata import Spec


class ModelOperation(Enum):
    """ Enumerate that describes the possible operations on Models
    """
    BEFORE_MODEL_RUN = 'before_model_run'
    SIMULATE = 'simulate'


class Model(metaclass=ABCMeta):
    """Abstract class represents the interface used to implement the model classes
    `SectorModel` and `ScenarioModel`.

    Arguments
    ---------
    name : str
    """

    def __init__(self, name):
        self.logger = getLogger(__name__)

        self.name = name
        self.description = ''
        self._inputs = {}
        self._parameters = {}
        self._outputs = {}

    def __repr__(self):
        return "<{} name='{}'>".format(self.__class__.__name__, self.name)

    @classmethod
    def from_dict(cls, config):
        """Create object from dictionary serialisation
        """
        model = cls(config['name'])
        model.description = config['description']
        for input_ in config['inputs']:
            model.add_input(Spec.from_dict(input_))
        for output in config['outputs']:
            model.add_output(Spec.from_dict(output))
        for param in config['parameters']:
            model.add_parameter(Spec.from_dict(param))
        return model

    def as_dict(self):
        """Serialize the SectorModel object as a dictionary

        Returns
        -------
        dict
        """
        config = {
            'name': self.name,
            'description': self.description,
            'path': sys.modules[self.__module__].__file__,
            'classname': self.__class__.__name__,
            'inputs': [inp.as_dict() for inp in self.inputs.values()],
            'outputs': [out.as_dict() for out in self.outputs.values()],
            'parameters': [param.as_dict() for param in self.parameters.values()]
        }
        return config

    @property
    def inputs(self):
        """All model inputs defined at this layer

        Returns
        -------
        dict of {input_name: smif.metadata.Spec}
        """
        return self._inputs

    @property
    def parameters(self):
        """Model parameters

        Returns
        -------
        dict of {parameter_name: smif.data_layer.data_array.DataArray}
        """
        return self._parameters

    def add_input(self, spec):
        """Add an input

        Arguments
        ---------
        spec: smif.metadata.Spec
        """
        self.inputs[spec.name] = spec

    def add_parameter(self, spec):
        """Add a parameter

        Arguments
        ---------
        spec: smif.metadata.Spec
        """
        self.parameters[spec.name] = spec

    @property
    def outputs(self):
        """All model outputs defined at this layer

        Returns
        -------
        dict of {output_name: smif.metadata.Spec}
        """
        return self._outputs

    def add_output(self, spec):
        """Add an output

        Arguments
        ---------
        spec: smif.metadata.Spec
        """
        self.outputs[spec.name] = spec

    @abstractmethod
    def simulate(self, data):
        """Override to implement the generation of model results

        Generate ``results`` for ``timestep`` using ``data``

        Arguments
        ---------
        data: smif.data_layer.DataHandle
            Access state, parameter values, dependency inputs.
        """


class ScenarioModel(Model):
    """Represents exogenous scenario data

    Arguments
    ---------
    name : str
        The name of this scenario (scenario set/abstract
        scenario/scenario group) - like sector model name

    Attributes
    ----------
    name : str
        Name of this scenario
    scenario : str
        Instance of scenario (concrete instance)
    """
    def __init__(self, name):
        super().__init__(name)
        self.scenario = None

    @classmethod
    def from_dict(cls, data):
        """Create ScenarioModel from dict serialisation
        """
        scenario = cls(data['name'])
        scenario.scenario = data['scenario']
        if 'description' in data:
            scenario.description = data['description']
        for output in data['outputs']:
            spec = Spec.from_dict(output)
            scenario.add_output(spec)

        return scenario

    def as_dict(self):
        """Serialise ScenarioModel to dict
        """
        config = {
            'name': self.name,
            'description': self.description,
            'scenario': self.scenario,
            'outputs': [
                output.as_dict()
                for output in self.outputs.values()
            ]
        }
        return config

    def simulate(self, data):
        """No-op, as the data is assumed to be already available in the store
        """
        return data



"""A config store holds the configuration data for running system-of-systems models with smif:
- model runs
- system-of-systems models
- model definitions
- strategies
- scenarios and scenario variants
- narratives
"""
from abc import ABCMeta, abstractmethod


class ConfigStore(metaclass=ABCMeta):
    """A ConfigStore must implement each of the abstract methods defined in this interface
    """
    # region Model runs
    @abstractmethod
    def read_model_runs(self):
        """Read all system-of-system model runs

        Returns
        -------
        list[~smif.controller.modelrun.ModelRun]
        """

    @abstractmethod
    def read_model_run(self, model_run_name):
        """Read a system-of-system model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        ~smif.controller.modelrun.ModelRun
        """

    @abstractmethod
    def write_model_run(self, model_run):
        """Write system-of-system model run

        Parameters
        ----------
        model_run : ~smif.controller.modelrun.ModelRun
        """

    @abstractmethod
    def update_model_run(self, model_run_name, model_run):
        """Update system-of-system model run

        Parameters
        ----------
        model_run_name : str
        model_run : ~smif.controller.modelrun.ModelRun
        """

    @abstractmethod
    def delete_model_run(self, model_run_name):
        """Delete a system-of-system model run

        Parameters
        ----------
        model_run_name : str
        """
    # endregion

    # region System-of-systems models
    @abstractmethod
    def read_sos_models(self):
        """Read all system-of-system models

        Returns
        -------
        list[~smif.model.sos_model.SosModel]
        """

    @abstractmethod
    def read_sos_model(self, sos_model_name):
        """Read a specific system-of-system model

        Parameters
        ----------
        sos_model_name : str

        Returns
        -------
        ~smif.model.sos_model.SosModel
        """

    @abstractmethod
    def write_sos_model(self, sos_model):
        """Write system-of-system model

        Parameters
        ----------
        sos_model : ~smif.model.sos_model.SosModel
        """

    @abstractmethod
    def update_sos_model(self, sos_model_name, sos_model):
        """Update system-of-system model

        Parameters
        ----------
        sos_model_name : str
        sos_model : ~smif.model.sos_model.SosModel
        """

    @abstractmethod
    def delete_sos_model(self, sos_model_name):
        """Delete a system-of-system model

        Parameters
        ----------
        sos_model_name : str
        """
    # endregion

    # region Models
    @abstractmethod
    def read_models(self):
        """Read all models

        Returns
        -------
        list[~smif.model.Model]
        """

    @abstractmethod
    def read_model(self, model_name):
        """Read a model

        Parameters
        ----------
        model_name : str

        Returns
        -------
        ~smif.model.Model
        """

    @abstractmethod
    def write_model(self, model):
        """Write a model

        Parameters
        ----------
        model : ~smif.model.Model
        """

    @abstractmethod
    def update_model(self, model_name, model):
        """Update a model

        Parameters
        ----------
        model_name : str
        model : ~smif.model.Model
        """

    @abstractmethod
    def delete_model(self, model_name):
        """Delete a model

        Parameters
        ----------
        model_name : str
        """
    # endregion

    # region Scenarios
    @abstractmethod
    def read_scenarios(self):
        """Read scenarios

        Returns
        -------
        list[~smif.model.ScenarioModel]
        """

    @abstractmethod
    def read_scenario(self, scenario_name):
        """Read a scenario

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        ~smif.model.ScenarioModel
        """

    @abstractmethod
    def write_scenario(self, scenario):
        """Write scenario

        Parameters
        ----------
        scenario : ~smif.model.ScenarioModel
        """

    @abstractmethod
    def update_scenario(self, scenario_name, scenario):
        """Update scenario

        Parameters
        ----------
        scenario_name : str
        scenario : ~smif.model.ScenarioModel
        """

    @abstractmethod
    def delete_scenario(self, scenario_name):
        """Delete scenario from project configuration

        Parameters
        ----------
        scenario_name : str
        """
    # endregion

    # region Scenario Variants
    @abstractmethod
    def read_scenario_variants(self, scenario_name):
        """Read variants of a given scenario

        Parameters
        ----------
        scenario_name : str

        Returns
        -------
        list[dict]
        """

    @abstractmethod
    def read_scenario_variant(self, scenario_name, variant_name):
        """Read a scenario variant

        Parameters
        ----------
        scenario_name : str
        variant_name : str

        Returns
        -------
        dict
        """

    @abstractmethod
    def write_scenario_variant(self, scenario_name, variant):
        """Write scenario to project configuration

        Parameters
        ----------
        scenario_name : str
        variant : dict
        """

    @abstractmethod
    def update_scenario_variant(self, scenario_name, variant_name, variant):
        """Update scenario to project configuration

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        variant : dict
        """

    @abstractmethod
    def delete_scenario_variant(self, scenario_name, variant_name):
        """Delete scenario from project configuration

        Parameters
        ----------
        scenario_name : str
        variant_name : str
        """
    # endregion

    # region Narratives
    @abstractmethod
    def read_narrative(self, sos_model_name, narrative_name):
        """Read narrative from sos_model

        Parameters
        ----------
        sos_model_name : str
        narrative_name : str
        """
    # endregion

    # region Strategies
    @abstractmethod
    def read_strategies(self, modelrun_name):
        """Read strategies for a given model run

        Parameters
        ----------
        model_run_name : str

        Returns
        -------
        list[dict]
        """

    @abstractmethod
    def write_strategies(self, modelrun_name, strategies):
        """Write strategies for a given model_run

        Parameters
        ----------
        model_run_name : str
        strategies : list[dict]
        """
    # endregion



"""This module contains implementations of :class:`~smif.adaptor.adaptor.Adaptor` to perform
conversions between units and between different spatial or temporal dimensions where sufficient
metadata is provided in the :class:`~smif.metadata.spec.Spec` definition.

These should be useful to link models in simple cases, where it may be reasonable to rely on
strong assumptions about the underlying distributions of the variables to be converted.
"""
from smif.convert.adaptor import Adaptor
from smif.convert.interval import IntervalAdaptor
from smif.convert.region import RegionAdaptor
from smif.convert.unit import UnitAdaptor

__all__ = ["Adaptor", "IntervalAdaptor", "UnitAdaptor", "RegionAdaptor"]

__author__ = "Will Usher, Tom Russell, Roald Schoenmakers"
__copyright__ = "Will Usher, Tom Russell, Roald Schoenmakers"
__license__ = "mit"



"""ModelLoader reads python modules as specified at runtime, loading and instantiating
objects.
"""
import importlib
import logging
import os


class ModelLoader(object):
    """Load Model from config

    Examples
    --------
    Call :py:meth:`ModelLoader.load` to create, load and return a
    :class:`Model` object.

    >>> loader = ModelLoader()
    >>> sector_model = loader.load(sector_model_config)
    >>> conversion_model = loader.load(conversion_model_config)
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def load(self, config):
        """Loads the model class specified by the config, returns an instance of that class
        using the Model.from_dict method.

        Arguments
        ---------
        config : dict
            The model configuration data. Must include:
                - name (name for smif internal use)
                - path (absolute path to python module file)
                - classname (name of Model implementation class)
                - anything required by the Model.from_dict classmethod

        Returns
        -------
        :class:`~smif.model.Model`
        """
        klass = self.load_model_class(config['name'], config['path'], config['classname'])
        if not hasattr(klass, 'from_dict'):
            msg = "Model '{}' does not have a ``from_dict`` method and " \
                  "cannot be loaded from config"
            raise KeyError(msg.format(config['name']))
        model_instance = klass.from_dict(config)
        if model_instance:
            return model_instance
        else:
            raise ValueError("Model not initialised from configuration data")

    def load_model_class(self, model_name, model_path, classname):
        """Dynamically load model class

        Arguments
        ---------
        model_name : str
            The name used internally to identify the SectorModel
        model_path : str
            The path to the python module which contains the SectorModel
            implementation
        classname : str
            The name of the class of the SectorModel implementation

        Returns
        -------
        class
            The SectorModel implementation
        """
        if not os.path.exists(model_path):
            msg = "Cannot find '{}' for the '{}' model".format(model_path, model_name)
            raise FileNotFoundError(msg)

        msg = "Importing model %s as class %s from module at %s"
        self.logger.info(msg, model_name, classname, model_path)

        spec = importlib.util.spec_from_file_location(model_name, model_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        klass = module.__dict__[classname]
        return klass



"""Test all ConfigStore implementations
"""
import getpass
import os
from copy import copy

import smif.data_layer.database.setup_database
from pytest import fixture, mark, param, raises
from smif.data_layer.database_interface import DbConfigStore
from smif.data_layer.file.file_config_store import YamlConfigStore
from smif.data_layer.memory_interface import MemoryConfigStore
from smif.exception import SmifDataExistsError, SmifDataNotFoundError

SKIP_ON_APPVEYOR = mark.skipif(
    'APPVEYOR' in os.environ and os.environ['APPVEYOR'],
    reason="Not yet set up with postgresql service on Appveyor CI")


# For testing, always use `test_smif` database; other connection details are configurable
DB_OPTIONS = {
    'dbname': 'test_smif'
}

# Try setting sane defaults if environment variables are not set
try:
    DB_OPTIONS['host'] = os.environ['PGHOST']
except KeyError:
    DB_OPTIONS['host'] = 'localhost'

try:
    DB_OPTIONS['port'] = os.environ['PGPORT']
except KeyError:
    DB_OPTIONS['port'] = 5432

try:
    DB_OPTIONS['user'] = os.environ['PGUSER']
except KeyError:
    DB_OPTIONS['user'] = getpass.getuser()

try:
    DB_OPTIONS['password'] = os.environ['PGPASSWORD']
except KeyError:
    DB_OPTIONS['password'] = ''


@fixture(
    params=[
        'memory',
        'text_file',
        # param('database', marks=SKIP_ON_APPVEYOR)
        param('database', marks=mark.skip)
    ])
def init_handler(request, setup_empty_folder_structure):
    if request.param == 'memory':
        handler = MemoryConfigStore()
    elif request.param == 'text_file':
        base_folder = setup_empty_folder_structure
        handler = YamlConfigStore(base_folder)
    elif request.param == 'database':
        # migrations dir
        migrations_dir = os.path.join(os.path.dirname(os.path.realpath(
            smif.data_layer.database.setup_database.__file__)), 'migrations')

        # run up migrations
        smif.data_layer.database.setup_database.up_migrations(
            migrations_dir, DB_OPTIONS['dbname'])

        # run down migrations with fixture teardown
        def teardown():
            smif.data_layer.database.setup_database.down_migrations(
                migrations_dir, DB_OPTIONS['dbname'])
        request.addfinalizer(teardown)

        handler = DbConfigStore(**DB_OPTIONS)

    return handler


@SKIP_ON_APPVEYOR
def test_db_connection():
    """Test that we can connect to a database in the test environment
    """
    store = DbConfigStore(**DB_OPTIONS)
    with store.database_connection.cursor() as cur:
        cur.execute('SELECT 1;')
        assert cur.fetchone() == (1,)


@fixture
def handler(init_handler, minimal_model_run, get_sos_model, get_sector_model_no_coords,
            strategies, scenario_no_coords):
    handler = init_handler

    # scenarios
    handler.write_scenario(scenario_no_coords)

    # models
    handler.write_model(get_sector_model_no_coords)

    # sos models
    handler.write_sos_model(get_sos_model)

    # model runs
    handler.write_model_run(minimal_model_run)

    # planning
    handler.write_strategies('test_modelrun', strategies)

    return handler


class TestModelRuns:
    """Read, write, update model runs
    """
    def test_read_model_runs(self, handler, minimal_model_run):
        actual = handler.read_model_runs()
        expected = [minimal_model_run]
        assert actual == expected

    def test_read_model_run(self, handler, minimal_model_run):
        assert handler.read_model_run('test_modelrun') == minimal_model_run

    def test_read_non_existing_model_run(self, handler):
        with raises(SmifDataNotFoundError):
            handler.read_model_run('non_existing')

    def test_write_model_run(self, handler, minimal_model_run):
        new_model_run = {
            'name': 'new_model_run_name',
            'description': 'Model run 2'
        }
        handler.write_model_run(new_model_run)
        actual = handler.read_model_runs()
        expected = [minimal_model_run, new_model_run]
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_write_existing_model_run(self, handler):
        with raises(SmifDataExistsError):
            handler.write_model_run(handler.read_model_run('test_modelrun'))

    def test_update_model_run(self, handler):
        updated_model_run = {
            'name': 'test_modelrun',
            'description': 'Model run'
        }
        handler.update_model_run('test_modelrun', updated_model_run)
        assert handler.read_model_runs() == [updated_model_run]

    def test_update_non_existing_model_run(self, handler):
        with raises(SmifDataNotFoundError):
            handler.update_model_run('non_existing', {'name': 'non_existing'})

    def test_delete_model_run(self, handler):
        handler.delete_model_run('test_modelrun')
        assert handler.read_model_runs() == []

    def test_delete_non_existing_model_run(self, handler):
        with raises(SmifDataNotFoundError):
            handler.delete_model_run('non_existing')


class TestSosModel:
    """Read, write, update, delete SosModel config
    """
    def test_read_sos_models(self, handler, get_sos_model):
        actual = handler.read_sos_models()
        expected = [get_sos_model]
        assert actual == expected

    def test_read_sos_model(self, handler, get_sos_model):
        assert handler.read_sos_model('energy') == get_sos_model

    def test_read_non_existing_sos_model(self, handler):
        with raises(SmifDataNotFoundError):
            handler.read_sos_model('non_existing')

    def test_write_sos_model(self, handler, get_sos_model):
        new_sos_model = copy(get_sos_model)
        new_sos_model['name'] = 'another_sos_model'
        handler.write_sos_model(new_sos_model)
        actual = handler.read_sos_models()
        expected = [get_sos_model, new_sos_model]
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_write_existing_sos_model(self, handler):
        with raises(SmifDataExistsError):
            handler.write_sos_model({'name': 'energy'})

    def test_update_sos_model(self, handler, get_sos_model):
        updated_sos_model = copy(get_sos_model)
        updated_sos_model['sector_models'] = ['energy_demand']
        handler.update_sos_model('energy', updated_sos_model)
        assert handler.read_sos_models() == [updated_sos_model]

    def test_update_non_existing_sos_model(self, handler):
        with raises(SmifDataNotFoundError):
            handler.update_sos_model('non_existing', {'name': 'non_existing'})

    def test_delete_sos_model(self, handler):
        handler.delete_sos_model('energy')
        assert handler.read_sos_models() == []

    def test_delete_non_existing_sos_model(self, handler):
        with raises(SmifDataNotFoundError):
            handler.delete_sos_model('non_existing')


class TestSectorModel():
    """Read/write/update/delete SectorModel config
    """
    def test_read_models(self, handler, get_sector_model_no_coords):
        actual = handler.read_models()
        expected = [get_sector_model_no_coords]
        assert actual == expected

    def test_read_model(self, handler, get_sector_model_no_coords):
        actual = handler.read_model(get_sector_model_no_coords['name'])
        expected = get_sector_model_no_coords
        assert actual == expected

    def test_read_non_existing_model(self, handler):
        with raises(SmifDataNotFoundError):
            handler.read_model('non_existing')

    def test_write_model(self, handler, get_sector_model_no_coords):
        new_sector_model = copy(get_sector_model_no_coords)
        new_sector_model['name'] = 'another_energy_sector_model'
        handler.write_model(new_sector_model)
        actual = handler.read_models()
        expected = [get_sector_model_no_coords, new_sector_model]
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_write_existing_model(self, handler, get_sector_model_no_coords):
        with raises(SmifDataExistsError):
            handler.write_model(handler.read_model(get_sector_model_no_coords['name']))

    def test_update_model(self, handler, get_sector_model_no_coords):
        name = get_sector_model_no_coords['name']
        expected = copy(get_sector_model_no_coords)
        expected['description'] = ['Updated description']
        handler.update_model(name, expected)
        actual = handler.read_model(name)
        assert actual == expected

    def test_update_non_existing_model(self, handler):
        with raises(SmifDataNotFoundError):
            handler.update_model('non_existing', {'name': 'non_existing'})

    def test_delete_model(self, handler, get_sector_model_no_coords):
        handler.delete_model(get_sector_model_no_coords['name'])
        expected = []
        actual = handler.read_models()
        assert actual == expected

    def test_delete_non_existing_model(self, handler):
        with raises(SmifDataNotFoundError):
            handler.delete_model('non_existing')


class TestStrategies():
    """Read strategies data
    """
    def test_read_strategies(self, handler, strategies):
        expected = strategies
        actual = handler.read_strategies('test_modelrun')
        assert sorted(actual, key=lambda d: d['description']) == \
            sorted(expected, key=lambda d: d['description'])


class TestScenarios():
    """Read and write scenario data
    """
    def test_read_scenarios(self, scenario_no_coords, handler):
        actual = handler.read_scenarios()
        assert actual == [scenario_no_coords]

    def test_read_scenario(self, scenario_no_coords, handler):
        actual = handler.read_scenario('mortality')
        assert actual == scenario_no_coords

    def test_read_non_existing_test_read_scenario(self, handler):
        with raises(SmifDataNotFoundError):
            handler.read_scenario('non_existing')

    def test_write_scenario(self, scenario_no_coords, handler):
        another_scenario = {
            'name': 'fertility',
            'description': 'Projected annual fertility rates',
            'variants': [],
            'provides': []
        }
        handler.write_scenario(another_scenario)
        actual = handler.read_scenario('fertility')
        expected = another_scenario
        assert actual == expected

    def test_write_existing_scenario(self, handler, get_sector_model_no_coords):
        with raises(SmifDataExistsError):
            handler.write_scenario(handler.read_scenario('mortality'))

    def test_update_scenario(self, scenario_no_coords, handler):
        another_scenario = {
            'name': 'mortality',
            'description': 'Projected annual mortality rates',
            'variants': [],
            'provides': []
        }
        handler.update_scenario('mortality', another_scenario)
        assert handler.read_scenarios() == [another_scenario]

    def test_update_non_existing_scenario(self, handler):
        with raises(SmifDataNotFoundError):
            handler.update_scenario('non_existing', {})

    def test_delete_scenario(self, handler):
        handler.delete_scenario('mortality')
        assert handler.read_scenarios() == []

    def test_delete_non_existing_scenario(self, handler):
        with raises(SmifDataNotFoundError):
            handler.delete_scenario('non_existing')

    def test_read_scenario_variants(self, handler, scenario_no_coords):
        actual = handler.read_scenario_variants('mortality')
        expected = scenario_no_coords['variants']
        assert actual == expected

    def test_read_scenario_variant(self, handler, scenario_no_coords):
        actual = handler.read_scenario_variant('mortality', 'low')
        expected = scenario_no_coords['variants'][0]
        assert actual == expected

    def test_write_scenario_variant(self, handler, scenario_no_coords):
        new_variant = {
            'name': 'high',
            'description': 'Mortality (High)',
            'data': {
                'mortality': 'mortality_high.csv'
            }
        }
        handler.write_scenario_variant('mortality', new_variant)
        actual = handler.read_scenario_variants('mortality')
        expected = [new_variant] + scenario_no_coords['variants']
        assert sorted_by_name(actual) == sorted_by_name(expected)

    def test_update_scenario_variant(self, handler):
        new_variant = {
            'name': 'low',
            'description': 'Mortality (Low)',
            'data': {
                'mortality': 'mortality_low.csv'
            }
        }
        handler.update_scenario_variant('mortality', 'low', new_variant)
        actual = handler.read_scenario_variants('mortality')
        expected = [new_variant]
        assert actual == expected

    def test_delete_scenario_variant(self, handler):
        handler.delete_scenario_variant('mortality', 'low')
        assert handler.read_scenario_variants('mortality') == []


def sorted_by_name(list_):
    """Helper to sort lists-of-dicts
    """
    return sorted(list_, key=lambda d: d['name'])
